// Copyright (C) 2005-2007 CEREA
//     Author(s): Irne Korsakissok
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/


#ifndef POLYPHEMUS_FILE_MODELS_PINGDRIVER_CXX


#include "PlumeInGridDriver.hxx"


namespace Polyphemus
{


 //! Main constructor.
  /*! Builds the driver and reads option keys in the configuration file.
    \param config_file configuration file.
  */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::PinGDriver(string config_file): BaseModel<T>(config_file),
    Model(config_file), config(config_file)
  {

    /*** Display options ***/

    config.SetSection("[display]");
    // Should iterations be displayed on screen?
    config.PeekValue("Show_iterations", option_display["show_iterations"]);
    // Should current date be displayed on screen?
    config.PeekValue("Show_date", option_display["show_date"]);

    config.SetSection("[domain]");
    config.PeekValue("Cartesian", option_cartesian);


    /*** Pointers to 3D data ***/

    this->D3_map["MeridionalWind_i"] = &MeridionalWind_i;
    this->D3_map["ZonalWind_i"] = &ZonalWind_i;
    this->D3_map["Temperature_i"] = &Temperature_i;

    this->D2_map["FirstLevelWind_i"] = &FirstLevelWind_i;
    this->D2_map["LowCloudiness_i"] = &LowCloudiness_i;
    this->D2_map["MediumCloudiness_i"] = &MediumCloudiness_i;
    this->D2_map["HighCloudiness_i"] = &HighCloudiness_i;
    this->D2_map["Insolation_i"] = &Insolation_i;

    /*** Major point sources emissions ***/
    
    config.SetSection("[data]");
    string data_file, emission_file;
    config.PeekValue("Data_description", data_file);
    ConfigStream data(data_file);
    data.SetSection("[plume-in-grid]");
    data.PeekValue("file_source", source_file);
    
    /*** Gaussian configuration ***/

    config.SetSection("[gaussian]");
    config.PeekValue("file_gaussian", gaussian_file);
  }


  //! Destructor.
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::~PinGDriver()
  {
  }

  //! Allocates memory.
  /*! Allocates grids and fields.
   */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
    ::Allocate()
  {
   
    /*** Winds ***/

    ZonalWind_i.Resize(this->GridZ3D, this->GridY3D, this->GridX3D_interf);
    ZonalWind_i_interp.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);
    FileZonalWind_i.Resize(this->GridZ3D, this->GridY3D,
			   this->GridX3D_interf);
    FileZonalWind_f.Resize(this->GridZ3D, this->GridY3D,
			   this->GridX3D_interf);

    MeridionalWind_i.Resize(this->GridZ3D, this->GridY3D_interf,
			    this->GridX3D);
    MeridionalWind_i_interp.Resize(this->GridZ3D, this->GridY3D,
				   this->GridX3D);
    FileMeridionalWind_i.Resize(this->GridZ3D, this->GridY3D_interf,
				this->GridX3D);
    FileMeridionalWind_f.Resize(this->GridZ3D, this->GridY3D_interf,
				this->GridX3D);

    /*** Temperature ***/

    Temperature_i.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);
    FileTemperature_i.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);
    FileTemperature_f.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);

    /*** First level wind module ***/

    FirstLevelWind_i.Resize(this->GridY2D, this->GridX2D);
    FileFirstLevelWind_i.Resize( this->GridY2D, this->GridX2D);
    FileFirstLevelWind_f.Resize( this->GridY2D, this->GridX2D);

    /*** Cloudiness ***/

    LowCloudiness_i.Resize( this->GridY2D, this->GridX2D);
    FileLowCloudiness_i.Resize( this->GridY2D, this->GridX2D);
    FileLowCloudiness_f.Resize( this->GridY2D, this->GridX2D);
    
    MediumCloudiness_i.Resize( this->GridY2D, this->GridX2D);
    FileMediumCloudiness_i.Resize( this->GridY2D, this->GridX2D);
    FileMediumCloudiness_f.Resize( this->GridY2D, this->GridX2D);
    
    HighCloudiness_i.Resize( this->GridY2D, this->GridX2D);
    FileHighCloudiness_i.Resize( this->GridY2D, this->GridX2D);
    FileHighCloudiness_f.Resize( this->GridY2D, this->GridX2D);

    /*** Insolation ***/

    Insolation_i.Resize( this->GridY2D, this->GridX2D);
    FileInsolation_i.Resize( this->GridY2D, this->GridX2D);
    FileInsolation_f.Resize( this->GridY2D, this->GridX2D);

  }


  //! Initialization.
  /*! Initializes the model and the output saver, and then performs the loop
    over all meteorological conditions with calls to the model (over all time
    steps) and to the output saver.
  */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::Init()
  {
    BaseModel<T>::Init();
    Allocate();

    /*** Initializations ***/

    Model.Init();
    OutputSaver.Init(Model);

    // Extracts simulation information from Eulerian model.
    delta_t_eul = Model.GetDelta_t();
    Nt_eul = Model.GetNt();
    date_min = Model.GetDate_min();

    // Is there a field "Altitude" to manage vertical levels?
    altitude_field = Model.HasField("Altitude");

    /***  Meteorological data ***/

    // The configuration-file path is the field "Data_description" in the main
    // configuration file.
    this->config.SetSection("[data]");
    string data_description_file = this->config.PeekValue("Data_description");
    ConfigStream data_description_stream(data_description_file);
    this->input_files["meteo"].Read(data_description_file, "meteo");
    this->input_files["gaussian_meteo"]
      .Read(data_description_file, "gaussian_meteo");

    /*** Winds ***/

    this->InitData("meteo", "ZonalWind", FileZonalWind_i, FileZonalWind_f,
		   this->current_date, ZonalWind_i);
    if (!option_cartesian)
      TransformZonalWind(ZonalWind_i);
    this->InitData("meteo", "MeridionalWind", FileMeridionalWind_i,
		   FileMeridionalWind_f, this->current_date,
		   MeridionalWind_i);
    if (!option_cartesian)
      TransformMeridionalWind(MeridionalWind_i);

    /*** Temperature ***/

    this->InitData("meteo", "Temperature", FileTemperature_i,
		   FileTemperature_f, this->current_date, Temperature_i);

    /*** FirstLevelWind ***/

    this->InitData("gaussian_meteo", "FirstLevelWindModule",
		   FileFirstLevelWind_i,FileFirstLevelWind_f,
		   this->current_date, FirstLevelWind_i);

    /*** LowCloudiness ***/

    this->InitData("gaussian_meteo", "LowCloudiness", FileLowCloudiness_i,
		   FileLowCloudiness_f, this->current_date, LowCloudiness_i);

    /*** MediumCloudiness ***/

    this->InitData("gaussian_meteo", "MediumCloudiness",
		   FileMediumCloudiness_i, FileMediumCloudiness_f,
		   this->current_date, MediumCloudiness_i);

    /*** HighCloudiness ***/

    this->InitData("gaussian_meteo", "HighCloudiness",
		   FileHighCloudiness_i, FileHighCloudiness_f,
		   this->current_date, HighCloudiness_i);

    /*** Insolation ***/

    this->InitData("gaussian_meteo", "SolarRadiation", FileInsolation_i,
		   FileInsolation_f, this->current_date, Insolation_i);


    /*** Loop over major point sources emissions ***/

    cout << "Reading major point sources emissions." << endl;
    ConfigStream source(source_file);

    GaussianPuffList.clear();
    Nsource = 0;
    string line;
    while (!source.IsEmpty())
      {
	line = source.GetLine();
	if (split(line)[0] == "[source]")
	  {
	    Date date_beg, date_end;
	    T lon, lat, height;
	    T release_time, release_duration, rate, abscissa, ordinate;
	    string species_name;
	    map<string, int> indices;
	    map<string, T> coord;

	    // Reading major point emission data.
	    source.PeekValue("Abscissa", lon);
	    source.PeekValue("Ordinate", lat);
	    source.PeekValue("Altitude", height);
	    date_beg = source.PeekValue("Date_beg");
	    date_end = source.PeekValue("Date_end");
	    source.PeekValue("Rate", rate);
	    source.PeekValue("Species", species_name);
	    species.push_back(species_name);

	    // List of source coordinates in Eulerian Grid.
	    int index_x, index_y, index_z;
	    GetCellIndices(lon, lat, height, index_z, index_y, index_x);
	    if (index_x < 0 || index_x >= this->Nx
		|| index_y < 0 || index_y >= this->Ny || index_z >= this->Nz)
	      throw string("Source coordinates are not inside")
		+ "simulation domain.";
	    indices["X"] = index_x;
	    indices["Y"] = index_y;
	    indices["Z"] = index_z;
	    coord["longitude"] = lon;
	    coord["latitude"] = lat;
	    coord["height"] = height;
	    source_indices.push_back(indices);
	    source_coordinates.push_back(coord);

	    // Conversion to cartesian coordinates.
	    const T pi = 3.14159265358979323846264;
	    const T earth_radius = 6371229.;
	    if (!option_cartesian)
	      {
		abscissa = earth_radius * lon;
		ordinate = earth_radius * sin(lat * pi / 180.);
	      }
	    else
	      {
		abscissa = lon;
		ordinate = lat;
	      }
	    // Time of release (seconds from beginning of the simulation).
	    release_time = date_beg.GetSecondsFrom(date_min);
	    // Puff duration (seconds).
	    release_duration = date_end.GetSecondsFrom(date_beg);
	    // Gaussian Model creation.
	    GaussianPuff<T>* PuffModel = new GaussianPuff<T>(gaussian_file);

	    // Gaussian model initialization.
	    PuffModel->Init();
	    PuffModel->SetDate(date_min);

	    // Sources initialization in Gaussian Puff model.
	    int Npuff;
	    T time_puff, puff_quantity;
	    Npuff = max(int(release_duration / delta_t_eul + 0.5), 1);

	    puff_quantity = rate * delta_t_eul;
	    quantity.push_back(puff_quantity);
	    for (int puff = 0; puff < Npuff; puff++)
	      {
		time_puff = release_time + puff * delta_t_eul;
		PuffModel->AddPuffSource(time_puff, puff_quantity, abscissa,
					ordinate, height, species_name);
	      }
	    PuffModel->InitPosition();
	    delta_t_local = PuffModel->GetDelta_t();
	    if (delta_t_local > delta_t_eul)
	      {
		cout << "Warning: local time step has been set equal"
		     << " to eulerian time step." << endl;
		delta_t_local = delta_t_eul;
	      }
	    Nt_local = int(delta_t_eul / delta_t_local);
	    GaussianPuffList.push_back(PuffModel);
	    Nsource++;
	  }
      }
  }


  //! Extracts meteorological data.
  /*! Extracts meteorological data from 3D fields at a given point.
    \param D3_map pointer to map containing 3D meteorological data.
    \param index_z cell index along z.
    \param index_y cell index along y.
    \param index_x cell index along x.
    \param met_data map containing meteorological data at the given point.
    \param stability stability class at the given point.
  */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::ExtractMeteo(map<string, Data<T, 3>* > D3_map,
		 map<string, Data<T, 2>* > D2_map,
		 int index_z, int index_y, int index_x,
		 T lon, T lat, Date date, map<string, T> &met_data,
		 string &stability, bool &isday)
  {
    /*** Base data for Gaussian models. ***/

    T temperature, wind_angle, wind, meridional, zonal;
    const T pi = 3.14159265358979323846264;

    //Interpolate meridional and zonal winds on cell centers.
    LinearInterpolationDimension(*D3_map["ZonalWind_i"],
				 ZonalWind_i_interp, 2);
    LinearInterpolationDimension(*D3_map["MeridionalWind_i"],
				 MeridionalWind_i_interp, 1);

    meridional = MeridionalWind_i_interp(index_z, index_y, index_x);
    zonal = ZonalWind_i_interp(index_z, index_y, index_x);
    wind_angle = atan2(zonal, meridional);
    wind = sqrt(zonal * zonal + meridional * meridional);
    temperature = (*D3_map["Temperature_i"])(index_z, index_y, index_x);
    met_data["temperature"] = temperature - 273.15;
    met_data["wind_angle"] = wind_angle * 180. / pi;
    met_data["wind"] = wind;
    met_data["wind_angle"] = wind_angle;

    /*** Stability class. ***/

    T surface_wind, solar_radiation;
    isday = IsDay(lon, lat, date);
    surface_wind = (*D2_map["FirstLevelWind_i"])(index_y, index_x);
    solar_radiation = (*D2_map["Insolation_i"])(index_y, index_x);
    
    // Computing total cloudiness.
    T cloudiness, low_cloudiness, medium_cloudiness, high_cloudiness;
    low_cloudiness = (*D2_map["LowCloudiness_i"])(index_y, index_x);
    medium_cloudiness = (*D2_map["MediumCloudiness_i"])(index_y, index_x);
    high_cloudiness = (*D2_map["HighCloudiness_i"])(index_y, index_x);
    cloudiness = ComputeTotalCloudiness<T, T>(low_cloudiness,
					      medium_cloudiness,
					      high_cloudiness);
    // Computing Pasquill stability class.
    stability = ComputePasquillStabilityClass(surface_wind, solar_radiation,
					      cloudiness, isday);
  }


  //! Model initialization for each step.
  /*! It reads on file the data that are is needed for the current step.
   */  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::InitStep()
  {
    Model.InitStep();
    
    /*** Winds ***/
    
    this->UpdateData("meteo", "ZonalWind", FileZonalWind_i,
		     FileZonalWind_f, ZonalWind_i);
    if (!option_cartesian)
      TransformZonalWind(ZonalWind_i);
    
    this->UpdateData("meteo", "MeridionalWind", FileMeridionalWind_i,
		     FileMeridionalWind_f, MeridionalWind_i);
    if (!option_cartesian)
      TransformMeridionalWind(MeridionalWind_i);
    
    /*** Temperature ***/
    
    this->UpdateData("meteo", "Temperature", FileTemperature_i,
		     FileTemperature_f, Temperature_i);
    
    /*** FirstLevelWind ***/
    
    this->UpdateData("gaussian_meteo", "FirstLevelWindModule",
		     FileFirstLevelWind_i, FileFirstLevelWind_f,
		     FirstLevelWind_i);
    
    /*** LowCloudiness ***/
    
    this->UpdateData("gaussian_meteo", "LowCloudiness", FileLowCloudiness_i,
		     FileLowCloudiness_f,  LowCloudiness_i);
    
    /*** MediumCloudiness ***/
    
    this->UpdateData("gaussian_meteo", "MediumCloudiness",
		     FileMediumCloudiness_i, FileMediumCloudiness_f,
		     MediumCloudiness_i);
    
    /*** HighCloudiness ***/
    
    this->UpdateData("gaussian_meteo", "HighCloudiness", FileHighCloudiness_i,
		     FileHighCloudiness_f, HighCloudiness_i);
    
    /*** Insolation ***/
    
    this->UpdateData("gaussian_meteo", "SolarRadiation", FileInsolation_i,
		     FileInsolation_f, Insolation_i);
    
    OutputSaver.InitStep(Model);
  }

  //! Performs the simulation.
  /*! Performs the loop over all time steps with calls to the model and to the
    output saver. For each major point emission, the corresponding Gaussian
    Model instance is called.
  */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::Run()
  {
    // Constant values.
    const T pi = 3.14159265358979323846264;
    const T earth_radius = 6371229.;

    /*** Time loop ***/

    for (int i = 0; i < Model.GetNt(); i++)
      {
	cout << endl;
	if (option_display["show_iterations"])
	  cout << "Performing iteration #" << i << endl;

	if (option_display["show_date"])
	  cout << "Current date: " <<
	    Model.GetCurrentDate().GetDate("%y-%m-%d %h:%i") << endl;
	InitStep();
	Model.Forward();

	for (int m = 0; m < Nsource; m++)
	  {
	    cout << "Processing source " << m << endl;
	    map<string, T> meteorological_data;
	    string stability;
	    bool isday;
	    list<Puff<T> > PuffList;
	    list<Puff<T> > PuffList_temp;
	    ExtractMeteo(this->D3_map, this->D2_map, source_indices[m]["Z"],
			 source_indices[m]["Y"], source_indices[m]["X"],
			 source_coordinates[m]["longitude"],
			 source_coordinates[m]["latitude"],
			 Model.GetCurrentDate(),
			 meteorological_data, stability, isday);
	    GaussianPuffList[m]->SetMeteo(meteorological_data["temperature"],
					   meteorological_data["wind_angle"],
					   meteorological_data["wind"],
					   stability);
	    GaussianPuffList[m]->SetDayNight(isday);
	    for (int j = 0; j < Nt_local; j++)
	      {
		GaussianPuffList[m]->InitStep();
		
		// Advection and diffusion for puffs.
		GaussianPuffList[m]->Advection();
		GaussianPuffList[m]->Diffusion();

		PuffList = GaussianPuffList[m]->GetPuffList();
		PuffList_temp.clear();
		for (typename list<Puff<T> >::iterator iter = PuffList.begin();
		     iter != PuffList.end(); iter++)
		  {
		    // Puff center coordinates.
		    T x_c, y_c, z_c, lon_c, lat_c;
		    x_c = iter->GetX();
		    y_c = iter->GetY();
		    z_c = iter->GetZ();
		    // Conversion to longitude/latitude (degrees).
		    if (!option_cartesian)
		      {
			lon_c = x_c / earth_radius;
			lat_c = asin(y_c / earth_radius) * 180. / pi;
		      }
		    else
		      {
			lon_c = x_c;
			lat_c = y_c;
		      }
		    // Gets corresponding cell in eulerian grid.
		    int index_x, index_y, index_z;
		    GetCellIndices(lon_c, lat_c, z_c,
				   index_z, index_y, index_x);

		    // Computing cell width along y.
		    T cell_width_z, cell_width_y, cell_width_x, cell_volume;
		    ComputeCellWidth(index_z, index_y, index_x,cell_width_z,
				     cell_width_y, cell_width_x, cell_volume);

		    // Tests if puff has to be transfered to Eulerian model.
		    T sigma_y = iter->GetHorizontalSigma();
		    if (index_x < 0 || index_x >= this->Nx
			|| index_y < 0 || index_y >= this->Ny
			|| index_z >= this->Nz)
		      {
			cout << "Puff reached the end of simulation domain."
			     << endl;
		      }
		    // Tests if puff size has reached cell width.
		    else if (sigma_y >= cell_width_y)
		      {
			T concentration = quantity[m] / cell_volume;
			int s = Model.GetSpeciesIndex(species[m]);
			Model.GetConcentration()
			  (s, index_z, index_y, index_x) += concentration;
		      }
		    else
		      PuffList_temp.push_back(*iter);
		  }
		// New puff list: transfered puffs have been removed.
		GaussianPuffList[m]->SetPuffList(PuffList_temp);
		GaussianPuffList[m]->AddTime(delta_t_local);
	      }
	  }
	OutputSaver.Save(Model);
	this->AddTime(delta_t_eul);
      }
  }


  //! Computes cell width (in meters) and cell volume at given indices.
  /*!
    \param k cell index along z.
    \param j cell index along y.
    \param i cell index along x.
    \param CellWidth_x (output) cell width along x in meters.
    \param CellWidth_y (output) cell width along y in meters.
    \param CellWidth_z (output) cell width along z in meters.
    \param CellVolume (output) cell volume in cubic meters.
   */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::ComputeCellWidth(int k, int j, int i, T& CellWidth_z,
		     T& CellWidth_y, T& CellWidth_x, T& CellVolume)
  {
    // Constant values.
    const T pi = 3.14159265358979323846264;
    const T earth_radius = 6371229.;

    Array<T, 1> GridY_interf = this->GridY3D_interf.GetArray();
    Array<T, 1> GridY_interf_trans(this->Ny + 1);

    for (int index_j = 0; index_j < this->Ny + 1; index_j++)
      GridY_interf_trans(index_j)
	= earth_radius * sin(GridY_interf(index_j) * pi / 180.);
    CellWidth_x = earth_radius * this->Delta_x * pi / 180.;
    CellWidth_y = GridY_interf_trans(j + 1) - GridY_interf_trans(j);
    if (altitude_field)
      CellWidth_z = Model.D3("Altitude")(k + 1, j, i)
	- Model.D3("Altitude")(k, j, i);
    else
      CellWidth_z = this->GridZ3D_interf(k + 1) - this->GridZ3D_interf(k);
    CellVolume = CellWidth_x * CellWidth_y * CellWidth_z;
  }


  //! Computes indices of the cell containing a given point in Eulerian grid.
  /*!
    \param lon longitude of the point (degrees).
    \param lat latitude of the point (degrees).
    \param height of the point (meters).
    \param k (output) cell index along z.
    \param j (output) cell index along y.
    \param i (output) cell index along x.
   */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::GetCellIndices(T lon, T lat, T height,
		   int& index_z, int& index_y, int& index_x)
  {
    index_x = int((lon - this->x_min) / this->Delta_x);
    index_y = int((lat - this->y_min) / this->Delta_y);
    if (altitude_field)
    for (int k=0; k < this->Nz; k++)
      {
	index_z = k;
	if (Model.D3("Altitude")(k+1, index_y, index_x) > height)
	  break;
      }
    else
      for (int k=0; k < this->Nz; k++)
	{
	  index_z = k;
	  if (this->GridZ3D_interf(k+1) > height)
	    break;
	}
  }
  /*! \brief Transforms the zonal wind to ease the numerical integration of
    advection. */
  /*! Formula: ZonalWind = ZonalWind / cos(latitude).
    \param ZonalWind (input/output) zonal wind.
    \note Coordinates associated with ZonalWind must be in degrees.
   */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::TransformZonalWind(Data<T, 3>& ZonalWind)
  {
    int i, j, k;
    
    const T pi(3.14159265358979323846264);
    const T ratio = pi / 180.;
    
    int Nx = ZonalWind.GetLength(2);
    int Ny = ZonalWind.GetLength(1);
    int Nz = ZonalWind.GetLength(0);
    
    for (k=0; k<Nz; k++)
      for (j=0; j<Ny; j++)
	for (i=0; i<Nx; i++)
	  ZonalWind(k, j, i) /= cos(ZonalWind[1].Value(k, j, i)
				    * ratio);
  }
  
  
  /*! \brief Transforms the meridional wind to ease the numerical integration
    of advection. */
  /*! Formula: MeridionalWind = MeridionalWind * cos(latitude).
    \param MeridionalWind (input/output) meridional wind.
    \note Coordinates associated with MeridionalWind must be in degrees.
   */
  template<class T, class ClassEulerianModel,
	   class ClassLocalModel, class ClassOutputSaver>
  void PinGDriver<T, ClassEulerianModel, ClassLocalModel, ClassOutputSaver>
  ::TransformMeridionalWind(Data<T, 3>& MeridionalWind)
  {
    
    int i, j, k;
    
    const T pi(3.14159265358979323846264);
    const T ratio = pi / 180.;
    
    int Nx = MeridionalWind.GetLength(2);
    int Ny = MeridionalWind.GetLength(1);
    int Nz = MeridionalWind.GetLength(0);
    
    for (k=0; k<Nz; k++)
      for (j=0; j<Ny; j++)
	for (i=0; i<Nx; i++)
	  MeridionalWind(k, j, i) *= cos(MeridionalWind[1].Value(k, j, i)
					 * ratio);
  }
  
} // namespace Polyphemus.


#define POLYPHEMUS_FILE_DRIVER_PINGDRIVER_CXX
#endif




 
