// Copyright (C) 2006-2007 CEREA
//     Author(s): Hadjira Foudhil, Irne Korsakissok
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/

// This file is part of the Gaussian models for Polyphemus.


#ifndef POLYPHEMUS_FILE_MODELS_BRIGGSFORMULA_HXX


namespace Polyphemus
{


  //! Estimates an effective source height of a hot source.
  /*! It uses the Briggs' plume rise formulation.
    \param temperature temperature of ambient air (K).
    \param wind wind velocity (m/s).
    \param inversion_height inversion height (m).
    \param stability stability in [0, 5].
    \param source_velocity efflux speed of gases (m/s).
    \param source_temperature temperature of emitted gases (K).
    \param diameter source diameter (m).
    \param source_height source height (m).
    \return The effective source height (m).
  */
  template<class T>
  T ComputeEffectiveSourceHeight(T temperature, T wind, T inversion_height,
				 int stability, T source_velocity,
				 T source_temperature, T diameter,
				 T source_height)
  {
    const T g = 9.81;

    T dTdz;

    // Vertical gradient of the potential temperature.
    if (stability == 0)
      dTdz = -0.015;
    else if (stability == 1)
      dTdz = -0.008;
    else if (stability == 2)
      dTdz = -0.006;
    else if (stability == 3)
      dTdz = 0.;
    else if (stability == 4)
      dTdz = 0.02;
    else if (stability == 5)
      dTdz = 0.035;
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";

    T sp_briggs;
    T Fm;
    T effective_height = 0.;
    T Dhz = 0.;
  

    // Briggs static stability parameter.
    sp_briggs = (g * dTdz)/temperature ;
  
    // Initial momentum flux parameter.
    Fm = g * source_velocity * 0.25 * diameter * diameter
      * abs(source_temperature - temperature) / source_temperature;
  
    if (dTdz > 0.)
      // Stable.
      Dhz = 2.6 * pow(Fm / (wind * sp_briggs), 1./3.);
    else
      // Unstable or neutral.
      Dhz = 38.71 / wind * pow(Fm, 3./5.);
  
    // Effective source height.
    effective_height = Dhz + source_height;

    if (inversion_height <= 0.)
      return effective_height;
    else
      return min(effective_height, inversion_height);
  }


  //! Computes horizontal diffusion parameter over rural area.
  /*! It uses Briggs' dispersion parameterizations for open country with
    Pasquill stability classes.
    \param distance distance downwind of the source (m).
    \param stability Pasquill stability classes in [0, 5].
    \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeRuralPlumeHorizontalSigma(T distance, int stability)
  {
    if (stability == 0)
      return 0.22 * distance / sqrt(1. + 1.e-4 * distance);
    else if (stability == 1)
      return 0.16 * distance / sqrt(1. + 1.e-4 * distance);
    else if (stability == 2)
      return 0.11 * distance / sqrt(1. + 1.e-4 * distance);
    else if (stability == 3)
      return 0.08 * distance / sqrt(1. + 1.e-4 * distance);
    else if (stability == 4)
      return 0.06 * distance / sqrt(1. + 1.e-4 * distance);
    else if (stability == 5)
      return 0.04 * distance / sqrt(1. + 1.e-4 * distance);
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";
  }


  //! Computes vertical diffusion parameter over rural area.
  /*! It uses Briggs' dispersion parameterizations for open country with
    Pasquill stability classes.
    \param distance distance downwind of the source (m).
    \param stability Pasquill stability classes in [0, 5].
    \return The vertical plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeRuralPlumeVerticalSigma(T distance, int stability)
  {
    if (stability == 0)
      return 0.2 * distance;
    else if (stability == 1)
      return 0.12 * distance;
    else if (stability == 2)
      return 0.08 * distance / sqrt(1. + 2.e-4 * distance);
    else if (stability == 3)
      return 0.06 * distance / sqrt(1. + 1.5e-3 * distance);
    else if (stability == 4)
      return 0.03 * distance / (1. + 3.e-4 * distance);
    else if (stability == 5)
      return 0.016 * distance / (1. + 3.e-4 * distance);
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";
  }


  //! Computes derivative of vertical diffusion parameter over rural area.
  /*! It uses Briggs' dispersion parameterizations for open country with
    Pasquill stability classes.
    \param distance distance downwind of the source (m).
    \param stability Pasquill stability classes in [0, 5].
    \return The derivative of vertical plume-dispersion parameter
    with respect to the distance downwind of the source.
  */
  template<class T>
  T DifferentiateRuralPlumeVerticalSigma(T distance, int stability)
  {
    if (stability == 0)
      return 0.2;
    else if (stability == 1)
      return 0.12;
    else if (stability == 2)
      {
	T diff_factor = sqrt(1. + 2.e-4 * distance) * (1. + 2.e-4 * distance);
	return 0.08 * (1. + 1.e-4 * distance) / diff_factor;
      }
    else if (stability == 3)
      {
	T diff_factor = sqrt(1. + 1.5e-3 * distance) *
	  (1. + 1.5e-3 * distance);
	return 0.03 * (2. + 1.5e-3 * distance) / diff_factor;
      }
    else if (stability == 4)
      {
	T diff_factor = (1. + 3.e-4 * distance) * (1. + 3.e-4 * distance);
	return 0.03 * (1. + 6.e-3 *  distance) / diff_factor;
      }
    else if (stability == 5)
      {
	T diff_factor = (1. + 3.e-4 * distance) * (1. + 3.e-4 * distance);
	return 0.016 * (1. + 6.e-3 *  distance) / diff_factor;
      }
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";
  }


  //! Computes horizontal diffusion parameter over urban area.
  /*! It uses Briggs' dispersion parameterizations for open country with
    Pasquill stability classes.
    \param distance distance downwind of the source (m).
    \param stability Pasquill stability classes in [0, 5].
    \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeUrbanPlumeHorizontalSigma(T distance, int stability)
  {
    if (stability == 0)
      return 0.32 * distance / sqrt(1. + 4.e-4 * distance);
    else if (stability == 1)
      return 0.32 * distance / sqrt(1. + 4.e-4 * distance);
    else if (stability == 2)
      return 0.22 * distance / sqrt(1. + 4.e-4 * distance);
    else if (stability == 3)
      return 0.16 * distance / sqrt(1. + 4.e-4 * distance);
    else if (stability == 4)
      return 0.11 * distance / sqrt(1. + 4.e-4 * distance);
    else if (stability == 5)
      return 0.11 * distance / sqrt(1. + 4.e-4 * distance);
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";
  }


  //! Computes vertical diffusion parameter over urban area.
  /*! It uses Briggs' dispersion parameterizations for open country with
    Pasquill stability classes.
    \param distance distance downwind of the source (m).
    \param stability Pasquill stability classes in [0, 5].
    \return The vertical plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeUrbanPlumeVerticalSigma(T distance, int stability)
  {
    if (stability == 0)
      return 0.24 * distance * sqrt(1. + 1.e-3 * distance);
    else if (stability == 1)
      return 0.24 * distance * sqrt(1. + 1.e-3 * distance);
    else if (stability == 2)
      return 0.2 * distance;
    else if (stability == 3)
      return 0.14 * distance / sqrt(1. + 3.e-4 * distance);
    else if (stability == 4)
      return 0.08 * distance /sqrt(1. + 1.5e-3 * distance);
    else if (stability == 5)
      return 0.08 * distance /sqrt(1. + 1.5e-3 * distance);
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";
  }


  //! Computes derivative of vertical diffusion parameter over urban area.
  /*! It uses Briggs' dispersion parameterizations for open country with
    Pasquill stability classes.
    \param distance distance downwind of the source (m).
    \param stability Pasquill stability classes in [0, 5].
    \return The derivative of vertical plume-dispersion parameter
    with respect to the distance downwind of the source.
  */
  template<class T>
  T DifferentiateUrbanPlumeVerticalSigma(T distance, int stability)
  {
    if (stability == 0 || stability == 1)
      return 0.12 * (2. + 3.e-3 * distance) / sqrt(1. + 1.e-3 * distance);
    else if (stability == 2)
      return 0.20;
    else if (stability == 3)
      {
	T diff_factor = sqrt(1. + 3e-4 * distance) * (1. + 3.e-4 * distance);
	return 0.07 * (2. + 3.e-4 * distance) / diff_factor;
      }
    else if (stability == 4 || stability == 5)
      {
	T diff_factor = sqrt(1. + 1.5e-3 * distance) *
	  (1. + 1.5e-3 * distance);
	return 0.04 * (2. + 1.5e-3 *  distance) / diff_factor;
      }
    else
      throw string("Stability index should be in [0, 5], but ")
	+ to_str(stability) + " was provided.";
  }


  /*! Computes horizontal diffusion parameter for situations of normal
  diffusion.
  */
/*! It uses Doury's dispersion parameterizations.
  \param t transfert time from source (s).
  \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeDouryNormalPlumeHorizontalSigma(T t)
  {
    if (t >= 0. && t < 240.)
      return pow(0.405 * t, 0.859);
    if (t >= 240. && t < 97000.)
      return pow(0.135 * t, 1.130);
    if (t >= 97000. && t < 508000.)
      return pow(0.463 * t, 1.000);
    if (t >= 508000. && t < 1300000.)
      return pow(6.5 * t, 0.824);
    else
      return pow(2.e+05 * t, 0.5);
  }


  /*! Computes vertical diffusion parameter for situations of normal
  diffusion.
  */
/*! It uses Doury's dispersion parameterizations.
  \param t transfert time from source (s).
  \return The vertical plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeDouryNormalPlumeVerticalSigma(T t)
  {
    if (t >= 0. && t < 240.)
      return pow(0.42 * t, 0.814);
    if (t >= 240. && t < 3280.)
      return pow(t, 0.685);
    else
      return sqrt(20. * t);
  }


  /*! Computes derivative pf vertical diffusion parameter for situations of
  normal diffusion.
  */
  /*! It uses Doury's dispersion parameterizations.
    \param t transfert time from source (s).
    \return The derivative of vertical plume-dispersion parameter (m)
    with respect to the distance from the source.
  */
  template<class T>
  T DifferentiateDouryNormalPlumeVerticalSigma(T t)
  {
    if (t >= 0. && t < 240.)
      return (0.42 * 0.814) * pow(0.42 * t, (0.814 - 1.));
    if (t >= 240. && t < 3280.)
      return 0.685 * pow(t, (0.685 - 1.));
    else
      return 10. / sqrt(20. * t);
 }


  //! Computes horizontal diffusion parameter for situations of low diffusion.
  /*! It uses Doury's dispersion parameterizations.
    \param t transfert time from source (s).
    \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeDouryLowPlumeHorizontalSigma(T t)
  {
    if (t >= 0. && t < 240.)
      return pow(0.405 * t, 0.859);
    if (t >= 240. && t < 97000.)
      return pow(0.135 * t, 1.130);
    if (t >= 97000. && t < 508000.)
      return pow(0.463 * t, 1.000);
    if (t >= 508000. && t < 1300000.)
      return pow(6.5 * t, 0.824);
    else
      return pow(2.e+05 * t, 0.5);
  }


  //! Computes vertical diffusion parameter for situations of low diffusion.
  /*! It uses Doury's dispersion parameterizations.
    \param t transfert time from source (s).
    \return The vertical plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeDouryLowPlumeVerticalSigma(T t)
  {
    return sqrt(20. * t);
  }


  /*! Computes derivative pf vertical diffusion parameter for situations of
  low diffusion.
  */
  /*! It uses Doury's dispersion parameterizations.  \param t
  transfert time from source (s).  \return The derivative of vertical
  plume-dispersion parameter (m) with respect to the distance from the source.
  */
  template<class T>
  T DifferentiateDouryLowPlumeVerticalSigma(T t)
  {
    return 10. / sqrt(20. * t);
  }

  //! Computes horizontal diffusion parameter with similarity theory.
  /*! It uses Weil (88) formula for the general form.
    \param t transfert time from source (s).
    \param z height (m).
    \param u_star friction velocity (m/s).
    \param L Monin-Obukhov length (m).
    \param h boundary layer height (m).
    \param f Coriolis parameter.
    \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeHorizontalSigma(T t, T z, T u_star, T h, T L, T f)
  {
    T sigma_v = ComputeWindHorizontalSigma(z, u_star, h, L, f);
    T tly = ComputeHorizontalTimeScale(z, u_star, h, L, f, sigma_v);
    T sigma_y = t *  sigma_v / sqrt(1 + 0.5 * t  / tly);
    return sigma_y;
  }


  //! Computes horizontal wind standard deviation.
  /*! It uses Hanna (82) formula.
    \param z height (m).
    \param u_star friction velocity (m/s).
    \param L Monin-Obukhov length (m).
    \param h boundary layer height (m).
    \param f Coriolis parameter.
    \return The horizontal wind standard deviation (m/s).
  */
  template<class T>
  T ComputeWindHorizontalSigma(T z, T u_star, T h, T L, T f)
  {
    // Unstable conditions.
    if (h / L < - 0.3)
      return u_star * pow((12. - 0.5 * h / L), 1./3.);
    // Stable conditions.
    else if (h / L > 1.)
      return 1.3 * u_star * (1. - z / h);
    // Neutral conditions.
    else
      return 1.3 * exp(-2. * f * z / u_star);
  }


  //! Computes Lagrangian time scale for horizontal dispersion.
  /*! It uses Hanna (84) formula.
    \param z height (m).
    \param h boundary layer height (m).
    \param L Monin-Obukhov length (m).
    \param f Coriolis parameter.
    \param sigma_v horizontal wind standard deviation (m/s).
    \return The Lagrangian time scale for horizontal dispersion (s).
  */
  template<class T>
  T ComputeHorizontalTimeScale(T z, T u_star, T h, T L, T f, T sigma_v)
  {
    // Unstable conditions.
    if (h / L < - 0.3)
      return 0.15 * h / sigma_v;
    // Stable conditions.
    else if (L > 1.)
      return 0.07 * h / sigma_v * sqrt(z / h);
    // Neutral conditions.
    else
      return 0.5 * z / sigma_v *(1 + 15 * f * z / u_star);
  }


  //! Computes vertical diffusion parameter with similarity theory.
  /*! It uses Weil (88) formula for the general form in unstable
    (convective) cases, and Irwin (79) for stable/neutral cases.
    \param t transfert time from source (s).
    \param z height (m).
    \param u_star friction velocity (m/s).
    \param w_star convective velocity (m/s).
    \param L Monin-Obukhov length (m).
    \param h boundary layer height (m).
    \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T ComputeVerticalSigma(T t, T z, T u_star, T w_star, T h, T L)
  {
    T sigma_w = ComputeWindVerticalSigma(z, u_star, w_star, h, L);
    T Fz;
    // Unstable conditions.
    if (h / L < - 0.3)
      {
	T tlz = ComputeVerticalTimeScale(z, h, sigma_w);
	T sigma_z = t *  sigma_w / sqrt(1 + 0.5 * t / tlz);
	return sigma_z;
      }
    // Neutral/Stable conditions.
    else
      {
	if (z < 50)
	  Fz = 1. / (1. + 0.9 * sqrt(t / 50.));
	else
	  Fz = 1. / (1. + 0.945 * pow( 0.1 * t, 0.806));
	T sigma_z = t * sigma_w * Fz;
	return sigma_z;
      }
  }

  /*! Computes derivative of the vertical diffusion parameter with similarity
    theory.*/
  /*! It uses Weil (88) formula for the general form in unstable
    (convective) cases, and Irwin (79) for stable/neutral cases.
    \param t transfert time from source (s).
    \param z height (m).
    \param u_star friction velocity (m/s).
    \param w_star convective velocity (m/s).
    \param L Monin-Obukhov length (m).
    \param h boundary layer height (m).
    \return The horizontal plume-dispersion parameter (m).
  */
  template<class T>
  T DifferentiateVerticalSigma(T t, T u, T z, T u_star, T w_star, T h, T L)
  {
    T sigma_w = ComputeWindVerticalSigma(z, u_star, w_star, h, L);
    T Fz, dFz;
    // Unstable conditions.
    if (h / L < - 0.3)
      {
	T tlz = ComputeVerticalTimeScale(z, h, sigma_w);
	T dfz = - 0.25 / (u * tlz * pow((1. + 0.5 * t / tlz), 3./2.));
	T fz = 1. / sqrt(1 + 0.5 * t / tlz);
	return sigma_w * t * dfz + sigma_w * fz / u;
      }
    // Neutral/Stable conditions.
    else
      {
	if (z < 50)
	  {
	    Fz = 1. / (1. + 0.9 * sqrt(t / 50.));
	    dFz = - Fz * Fz * 0.09 / (u * sqrt(t / 50.));
	  }
	else
	  {
	    Fz = 1. / (1. + 0.945 * pow( 0.1 * t, 0.806));
	    dFz = - Fz * Fz * 0.762 / (100. * u * pow((t / 100), 0.194));
	  }
	return t * sigma_w * dFz + sigma_w * Fz / u;
      }
  }


  //! Computes vertical wind standard deviation.
  /*! It uses Weil (88) formula for convective cases
    and Venkatram (84) for stable/neutral cases.
    \param z height (m).
    \param u_star friction velocity (m/s).
    \param L Monin-Obukhov length (m).
    \param h boundary layer height (m).
    \param f Coriolis parameter.
    \return The vertical wind standard deviation (m/s).
  */
  template<class T>
  T ComputeWindVerticalSigma(T z, T u_star, T w_star, T h, T L)
  {
    // Unstable conditions.
    if (h / L < - 0.3)
      return 0.6 * w_star;
    // Neutral/Stable conditions.
    else
      return 1.3 * u_star * pow((1. - z / h), 3./4.);
  }


  /*! Computes Lagrangian time scale for vertical dispersion in unstable
    cases.*/
  /*! It uses Hanna (82) formula.
    \param z height (m).
    \param h boundary layer height (m).
    \param sigma_w vertical wind standard deviation (m/s).
    \return The Lagrangian time scale for vertical dispersion (s).
  */
  template<class T>
  T ComputeVerticalTimeScale(T z, T h, T sigma_w)
  {
    return 0.15 * h * (1. - exp(-5. * z / h)) / sigma_w;
  }


  //! Computes a concentration at a given point.
  /*! Gaussian plume model for a continuous point source in uniform flow with
    homogeneous turbulence.
    \param wind mean transport velocity across the plume (m/s).
    \param inversion_height inversion height (m).
    \param effective_height effective source height (m).
    \param rate source strength or emission rate (mass unit/s).
    \param y ordinate (crosswind) (m).
    \param z height (m).
    \param sigma_y horizontal diffusion parameter (m).
    \param sigma_z vertical diffusion parameter (m).
    \param alpha coefficient for Overcamp model (in [-1, 1])..
    \param minimum_volume (optional) minimum volume in which the tracer is
    mixed within one second (m^3/s). It is used to set an upper bound on the
    concentration. Put zero to avoid this threshold. Default: 1.e-6.
    \return The concentration at (\a y, \a z).
    \warning The wind should not be zero.
  */
  template<class T>
  T ComputePlumeConcentration(T wind, T inversion_height, T effective_height,
			      T rate, T y, T z, T sigma_y, T sigma_z, T alpha,
			      T minimum_volume = 1.e-6)
  {
    const T pi = 3.14159265358979323846264;

    // Crosswind contribution.
    T Fy = 0.;
    Fy = exp(-y*y  / (2. * sigma_y*sigma_y));
  
    // Vertical contribution (real and image sources respectively).
    T Fz = 0.;
    T dzr = z - effective_height;
    T dzi = z + effective_height;

    T sigma_z_2 = sigma_z*sigma_z;
  
    Fz = exp(-dzr*dzr / (2. * sigma_z_2))
      +  alpha * exp(-dzi*dzi / (2. * sigma_z_2));

    // Reflexion if inversion (real and image sources respectively).
    T FzI = 0.;

    if (inversion_height > 0.)
      for (int N = 1; N < 6; N++)
	{
	  T dzmi = z - effective_height + 2. * N * inversion_height;
	  T dzpi = z + effective_height - 2. * N * inversion_height;
	  
	  FzI += exp(-dzmi*dzmi / (2. * sigma_z_2))
	    + alpha * exp(-dzpi*dzpi / (2. * sigma_z_2));
	}

    T concentration
      = rate * Fy * (Fz + FzI) / (2. * pi * wind * sigma_y * sigma_z);

    return minimum_volume != 0. ? min(concentration, rate / minimum_volume)
      : concentration;
  }


}


#define POLYPHEMUS_FILE_MODELS_BRIGGSFORMULA_HXX
#endif
