// Copyright (C) 2006-2007 CEREA
//     Author(s): Vivien Mallet
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the air quality modeling system Polyphemus.
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/

// This file is part of the Eulerian model Polair3D.

#ifndef POLYPHEMUS_FILE_MODELS_POLAIR3DAEROSOL_CXX


#include "Polair3DAerosol.hxx"


namespace Polyphemus
{


  ////////////////////////////////
  // CONSTRUCTOR AND DESTRUCTOR //
  ////////////////////////////////


  //! Main constructor.
  /*!
    \param config_file configuration file.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Polair3DAerosol(string config_file):
    Polair3DChemistry<T, ClassAdvection, ClassDiffusion,
		      ClassChemistry>(config_file)
  {

    /*** Managed data ***/

    // Pressure and temperature are only needed for air density
    // or microphysical model of scavenging.
    this->option_manage["temperature"] =
      this->option_process["with_air_density"]
      || this->scavenging_model == "microphysical"
      || this->scavenging_model == "microphysical-ph";
    this->option_manage["pressure"] =
      this->option_process["with_air_density"]
      || this->scavenging_model == "microphysical"
      || this->scavenging_model == "microphysical-ph";

    this->option_manage["initial_condition_aer"] = true;
    this->option_manage["surface_temperature"] = true;
    this->option_manage["surface_pressure"] = true;
    this->option_manage["first_level_wind_module"] = true;
    this->option_manage["liquid_water_content"] = true;
    this->option_manage["snow_height"] = true;
    this->option_manage["boundary_condition_aer"] = true;
    this->option_manage["deposition_velocity_aer"] = true;
    this->option_manage["scavenging_coefficient_aer"] = true;
    this->option_manage["surface_emission_aer"] = true;
    this->option_manage["volume_emission_aer"] = true;

    /*** Pointers to 2D data ***/

    this->D2_map["SurfaceTemperature"] = &SurfaceTemperature_f;
    this->D2_map["SurfaceTemperature_i"] = &SurfaceTemperature_i;
    this->D2_map["SurfaceTemperature_f"] = &SurfaceTemperature_f;

    this->D2_map["SurfacePressure"] = &SurfacePressure_f;
    this->D2_map["SurfacePressure_i"] = &SurfacePressure_i;
    this->D2_map["SurfacePressure_f"] = &SurfacePressure_f;

    this->D2_map["FirstLevelWindModule"] = &FirstLevelWindModule_f;
    this->D2_map["FirstLevelWindModule_i"] = &FirstLevelWindModule_i;
    this->D2_map["FirstLevelWindModule_f"] = &FirstLevelWindModule_f;

    this->D2_map["SnowHeight"] = &SnowHeight_f;
    this->D2_map["SnowHeight_i"] = &SnowHeight_i;
    this->D2_map["SnowHeight_f"] = &SnowHeight_f;

    /*** Pointers to 3D data ***/

    this->D3_map["LiquidWaterContent"] = &LiquidWaterContent_i;
    this->D3_map["LiquidWaterContent_i"] = &LiquidWaterContent_i;

    this->D3_map["DepositionVelocity_aer"] = &DepositionVelocity_aer_f;
    this->D3_map["DepositionVelocity_aer_i"] = &DepositionVelocity_aer_i;
    this->D3_map["DepositionVelocity_aer_f"] = &DepositionVelocity_aer_f;

    this->D3_map["pH"] = &pH;

    /*** Pointers to 4D data ***/

    this->D4_map["BoundaryCondition_z_aer"] = &BoundaryCondition_z_aer_i;
    this->D4_map["BoundaryCondition_z_aer_i"] = &BoundaryCondition_z_aer_i;

    this->D4_map["ScavengingCoefficient_aer"] = &ScavengingCoefficient_aer_f;
    this->D4_map["ScavengingCoefficient_aer_i"]
      = &ScavengingCoefficient_aer_i;
    this->D4_map["ScavengingCoefficient_aer_f"]
      = &ScavengingCoefficient_aer_f;

    this->D4_map["SurfaceEmission_aer"] = &SurfaceEmission_aer_f;
    this->D4_map["SurfaceEmission_aer_i"] = &SurfaceEmission_aer_i;
    this->D4_map["SurfaceEmission_aer_f"] = &SurfaceEmission_aer_f;

    this->D4_map["WetDiameter_aer"] = &WetDiameter_aer;

    this->D4_map["DryDepositionFlux_aer"] = &DryDepositionFlux_aer;
    this->D4_map["WetDepositionFlux_aer"] = &WetDepositionFlux_aer;
    this->D4_map["InCloudWetDepositionFlux_aer"]
      = &InCloudWetDepositionFlux_aer;

    /*** Pointers to 5D data ***/

    this->D5_map["BoundaryCondition_y_aer"] = &BoundaryCondition_y_aer_i;
    this->D5_map["BoundaryCondition_y_aer_i"] = &BoundaryCondition_y_aer_i;

    this->D5_map["BoundaryCondition_x_aer"] = &BoundaryCondition_x_aer_i;
    this->D5_map["BoundaryCondition_x_aer_i"] = &BoundaryCondition_x_aer_i;

    this->D5_map["VolumeEmission_aer"] = &VolumeEmission_aer_f;
    this->D5_map["VolumeEmission_aer_i"] = &VolumeEmission_aer_i;
    this->D5_map["VolumeEmission_aer_f"] = &VolumeEmission_aer_f;

    this->D5_map["Source_aer"] = &Source_aer_f;
    this->D5_map["Source_aer_i"] = &Source_aer_i;
    this->D5_map["Source_aer_f"] = &Source_aer_f;

    /*** Fields bins lists ***/

    this->field_bins["DepositionVelocity_aer"] = &bin_list_dep_aer;
    this->field_bins["ScavengingCoefficient_aer"] = &bin_list_scav_aer;
  }

  
  //! Destructor.
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::~Polair3DAerosol()
  {
  }


  ///////////////////
  // CONFIGURATION //
  ///////////////////

  
  //! Reads the configuration.
  /*! It reads the description of the domain, the simulation starting-date,
    species lists, options (especially which processes are included) and the
    paths to data input-files.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::ReadConfiguration()
  {
    Polair3DChemistry<T, ClassAdvection, ClassDiffusion,
      ClassChemistry>::ReadConfiguration();

    /*** Options ***/

    this->config.SetSection("[options]");
    this->config.PeekValue("With_initial_condition_aerosol",
			   this->option_process
			   ["with_initial_condition_aer"]);
    this->config.PeekValue("With_boundary_condition_aerosol",
			   this->option_process
			   ["with_boundary_condition_aer"]);
    this->config.PeekValue("With_deposition_aerosol",
			   this->option_process["with_deposition_aer"]);
    if (this->option_process["with_deposition_aer"])
      this->config.PeekValue("Compute_deposition_aerosol",
			     this->option_process["compute_deposition_aer"]);
    else
      this->option_process["compute_deposition_aer"] = false;
    this->config.PeekValue("With_scavenging_aerosol",
			   this->option_process["with_scavenging_aer"]);
    this->config.PeekValue("With_in_cloud_scavenging",
			   this->option_process["with_in_cloud_scavenging"]);
    this->config.PeekValue("With_point_emission_aerosol",
			   this->option_process["with_point_emission_aer"]);
    this->config.PeekValue("With_surface_emission_aerosol",
			   this->option_process["with_surface_emission_aer"]);
    this->config.PeekValue("With_volume_emission_aerosol",
			   this->option_process["with_volume_emission_aer"]);
    this->config.PeekValue("Collect_dry_flux_aerosol",
			   this->option_process["collect_dry_flux_aer"]);
    this->config.PeekValue("Collect_wet_flux_aerosol",
			   this->option_process["collect_wet_flux_aer"]);
    // Warning: should be set to true only if chemistry computes pH.
    this->config.PeekValue("With_pH", this->option_process["with_pH"]);

    // Liquid water content threshold above which there are clouds (in g/m3).
    this->config.PeekValue("Lwc_cloud_threshold", lwc_cloud_threshold);
    // Density in kg / m^3.
    this->config.PeekValue("Fixed_aerosol_density", fixed_density_aer);

    /*** Disables management of useless fields ***/

    // Variables only needed for computing dry deposition.
    this->option_manage["surface_temperature"] =
      this->option_manage["surface_temperature"]
      && this->option_process["compute_deposition_aer"];
    this->option_manage["surface_pressure"] =
      this->option_manage["surface_pressure"]
      && this->option_process["compute_deposition_aer"];
    this->option_manage["first_level_wind_module"] =
      this->option_manage["first_level_wind_module"]
      && this->option_process["compute_deposition_aer"];
    this->option_manage["snow_height"] = this->option_manage["snow_height"]
      && this->option_process["compute_deposition_aer"];
    this->option_manage["specific_humidity"] = true;

    // Rain and cloud height are only needed for scavenging.
    this->option_manage["rain"] = this->scavenging_model != "none"
      || this->option_process["with_scavenging_aer"]
      || this->option_process["with_in_cloud_scavenging"];
    this->option_manage["cloud_height"] = this->scavenging_model != "none"
      || this->option_process["with_scavenging_aer"];

    /*** Species and bins ***/

    // Opens the file that describes species.
    ConfigStream species_stream(this->file_species);
    // Section "[aerosol_species]" contains all aerosol species names.
    species_stream.SetSection("[aerosol_species]");
    while (!species_stream.IsEmpty())
      this->species_list_aer.push_back(species_stream.GetElement());
    this->Ns_aer = int(this->species_list_aer.size());
    
    // Reads bin bounds.
    this->config.SetSection("[domain]");
    this->config.Find("Bin_bounds");
    bin_list = split(this->config.GetLine());
    this->Nbin_aer = int(bin_list.size()) - 1;

    /*** Input files ***/
    
    vector<string> species_bin;
    string species;
    int bin_index;

    // The configuration-file path is the field "Data_description" in the main
    // configuration file.
    this->config.SetSection("[data]");
    string data_description_file = this->config.PeekValue("Data_description");
    // Opens the configuration file for input data.
    ConfigStream data_description_stream(data_description_file);

    // Initial conditions.
    if (this->option_process["with_initial_condition_aer"])
      this->input_files["initial_condition_aer"]
	.ReadFiles(data_description_file, "initial_condition_aerosol");
    else
      this->input_files["initial_condition_aer"].Empty();
    for (map<string, string>::iterator i
	   = this->input_files["initial_condition_aer"].Begin();
	 i != this->input_files["initial_condition_aer"].End(); i++)
      {
	species_bin = split(i->first, "_");
	if (species_bin.size() != 2)
	  throw string("Species \"") + i->first + "\" is badly formatted.";
	species = species_bin[0];
	bin_index = convert<int>(species_bin[1]);
	unsigned int j = 0;
	while (j < species_list_ic_aer.size()
	       && species_list_ic_aer[j].first != species)
	  j++;
	if (j == species_list_ic_aer.size())
	  species_list_ic_aer
	    .push_back(pair<string, vector<int> >(species, vector<int>()));
	species_list_ic_aer[j].second.push_back(bin_index);
      }
    Ns_ic_aer = int(species_list_ic_aer.size());

    // Boundary conditions.
    if (this->option_process["with_boundary_condition_aer"])
      this->input_files["boundary_condition_aer"]
	.Read(data_description_file, "boundary_condition_aerosol");
    else
      this->input_files["boundary_condition_aer"].Empty();
    Nbin_bc_aer = 0;
    for (map<string, string>::iterator i
	   = this->input_files["boundary_condition_aer"].Begin();
	 i != this->input_files["boundary_condition_aer"].End(); i++)
      {
	species_bin = split(i->first, "_");
	if (species_bin.size() != 2)
	  throw string("Species \"") + i->first + "\" is badly formatted.";
	species = species_bin[0];
	bin_index = convert<int>(species_bin[1]);
	unsigned int j = 0;
	while (j < species_list_bc_aer.size()
	       && species_list_bc_aer[j].first != species)
	  j++;
	if (j == species_list_bc_aer.size())
	  species_list_bc_aer
	    .push_back(pair<string, vector<int> >(species, vector<int>()));
	species_list_bc_aer[j].second.push_back(bin_index);
	Nbin_bc_aer
	  = max(Nbin_bc_aer, int(species_list_bc_aer[j].second.size()));
      }
    Ns_bc_aer = int(species_list_bc_aer.size());

    // Deposition velocities.
    if (this->option_process["with_deposition_aer"])
      if (this->option_process["compute_deposition_aer"])
	this->input_files["deposition_velocity_aer"]
	  .ReadFields(data_description_file, "deposition_velocity_aerosol");
      else
	this->input_files["deposition_velocity_aer"]
	  .Read(data_description_file, "deposition_velocity_aerosol");
    else
      this->input_files["deposition_velocity_aer"].Empty();
    for (map<string, string>::iterator i
	   = this->input_files["deposition_velocity_aer"].Begin();
	 i != this->input_files["deposition_velocity_aer"].End(); i++)
      {
	if (!is_integer(i->first))
	  throw string("In deposition fields, \"") + i->first
	    + "\" is not a bin index.";
	bin_index = convert<int>(i->first);
	bin_list_dep_aer.push_back(bin_index);
      }
    Nbin_dep_aer = int(bin_list_dep_aer.size());

    // Scavenging for gaseous species.
    if (this->scavenging_model == "microphysical-ph")
      {
	// Henry constants in mol / L / atm.
	species_stream.SetSection("[henry]");
	while (!species_stream.IsEmpty())
	  {
	    species = species_stream.GetElement();
	    species_stream.GetNumber(this->henry_constant[species]);
	  }

	// Gas phase diffusivity constants in cm^2 / s.
	species_stream.SetSection("[diffusivity]");
	while (!species_stream.IsEmpty())
	  {
	    species = species_stream.GetElement();
	    species_stream.GetNumber(this->gas_phase_diffusivity[species]);
	  }

	// Dissolution heat (kcal / mol at 298K).
	species_stream.SetSection("[dissolution_heat]");
	while (!species_stream.IsEmpty())
	  {
	    species = species_stream.GetElement();
	    species_stream.GetNumber(dissolution_heat[species]);
	  }
      }
    
    // Scavenging for aerosols.

    if (this->option_process["with_scavenging_aer"])
      {
	this->option_manage["rain"] = true;
	this->option_manage["cloud_height"] = true;
	this->input_files["scavenging_aer"]
	  .ReadFields(data_description_file, "scavenging_aerosol");
      }
    else
      this->input_files["scavenging_aer"].Empty();

    for (map<string, string>::iterator i
	   = this->input_files["scavenging_aer"].Begin();
	 i != this->input_files["scavenging_aer"].End(); i++)
      {
	if (!is_integer(i->first))
	  throw string("In scavenging fields, \"") + i->first
	    + "\" is not a bin index.";
	bin_index = convert<int>(i->first);
	bin_list_scav_aer.push_back(bin_index);
      }
    Nbin_scav_aer = int(bin_list_scav_aer.size());

    // Point emissions.
    if (this->option_process["with_point_emission_aer"])
      {
	string point_emission_file;
	data_description_stream.SetSection("[point_emission_aerosol]");
	data_description_stream.PeekValue("file", point_emission_file);
	ConfigStream point_emission_stream(point_emission_file);

	// Parses the file that describes all point emissions.
	string type, str, line;
	vector<string> str_split;
	T x, y, z;
	int index;
	int Nsource = 0;
	while (!point_emission_stream.IsEmpty())
	  {
	    line = point_emission_stream.GetLine();
	    // A new source should be added.
	    if (split(line)[0] == "[source]")
	      {
		point_emission_list_aer.push_back(map<string, string>());
		// Emitted species.
		point_emission_stream.PeekValue("Species", str);
		str_split = split(str, "_");
		if (str_split.size() != 2)
		  throw string("In point emissions, badly formatted ")
		    + string("species: \"") + str + "\".";
		point_emission_list_aer[Nsource]["species"] = str_split[0];
		point_emission_list_aer[Nsource]["bin"] = str_split[1];
		// Coordinates.
		point_emission_stream.PeekValue("Abscissa", x);
		point_emission_stream.PeekValue("Ordinate", y);
		point_emission_stream.PeekValue("Altitude", z);
		// Maps its coordinates to model grid indices.
		index = int((x - this->x_min + this->Delta_x / 2.)
			    / this->Delta_x);
		point_emission_list_aer[Nsource]["i"] = to_str(index);
		index = int((y - this->y_min + this->Delta_y / 2.)
			    / this->Delta_y);
		point_emission_list_aer[Nsource]["j"] = to_str(index);
		index = 0;
		while (index < this->Nz && z > this->LayerInterface(index+1))
		  index++;
		point_emission_list_aer[Nsource]["k"] = to_str(index);
		// Source type (instantaneous or continuous).
		point_emission_stream.PeekValue("Type", type);
		point_emission_list_aer[Nsource]["type"] = type;
		if (type == "puff")
		  // Instantaneous emission.
		  {
		    point_emission_stream.PeekValue("Date", str);
		    point_emission_list_aer[Nsource]["date"] = str;
		    // Quantity emitted.
		    point_emission_stream.PeekValue("Quantity", str);
		    point_emission_list_aer[Nsource]["quantity"] = str;
		  }
		else if (type == "continuous")
		  // Continuous emission.
		  {
		    point_emission_stream.PeekValue("Date_beg", str);
		    point_emission_list_aer[Nsource]["date_beg"] = str;
		    point_emission_stream.PeekValue("Date_end", str);
		    point_emission_list_aer[Nsource]["date_end"] = str;
		    // Emission rate.
		    point_emission_stream.PeekValue("Rate", str);
		    point_emission_list_aer[Nsource]["rate"] = str;
		  }
		else
		  throw string("Unknown type of point emission: \"")
		    + type + "\".";
		Nsource++;
	      }
	  }
      }

    // Surface emissions.
    if (this->option_process["with_surface_emission_aer"])
      this->input_files["surface_emission_aer"]
	.Read(data_description_file, "surface_emission_aerosol");
    else
      this->input_files["surface_emission_aer"].Empty();
    Nbin_surf_emis_aer = 0;
    for (map<string, string>::iterator i
	   = this->input_files["surface_emission_aer"].Begin();
	 i != this->input_files["surface_emission_aer"].End(); i++)
      {
	species_bin = split(i->first, "_");
	if (species_bin.size() != 2)
	  throw string("Species \"") + i->first + "\" is badly formatted.";
	species = species_bin[0];
	bin_index = convert<int>(species_bin[1]);
	unsigned int j = 0;
	while (j < species_list_surf_emis_aer.size()
	       && species_list_surf_emis_aer[j].first != species)
	  j++;
	if (j == species_list_surf_emis_aer.size())
	  species_list_surf_emis_aer
	    .push_back(pair<string, vector<int> >(species, vector<int>()));
	species_list_surf_emis_aer[j].second.push_back(bin_index);
	Nbin_surf_emis_aer =
	  max(Nbin_surf_emis_aer,
	      int(species_list_surf_emis_aer[j].second.size()));
      }
    Ns_surf_emis_aer = int(species_list_surf_emis_aer.size());

    // Volume emissions.
    if (this->option_process["with_volume_emission_aer"])
      this->input_files["volume_emission_aer"]
	.Read(data_description_file, "volume_emission_aerosol");
    else
      this->input_files["volume_emission_aer"].Empty();
    Nbin_vol_emis_aer = 0;
    for (map<string, string>::iterator i
	   = this->input_files["volume_emission_aer"].Begin();
	 i != this->input_files["volume_emission_aer"].End(); i++)
      {
	species_bin = split(i->first, "_");
	if (species_bin.size() != 2)
	  throw string("Species \"") + i->first + "\" is badly formatted.";
	species = species_bin[0];
	bin_index = convert<int>(species_bin[1]);
	unsigned int j = 0;
	while (j < species_list_vol_emis_aer.size()
	       && species_list_vol_emis_aer[j].first != species)
	  j++;
	if (j == species_list_vol_emis_aer.size())
	  species_list_vol_emis_aer
	    .push_back(pair<string, vector<int> >(species, vector<int>()));
	species_list_vol_emis_aer[j].second.push_back(bin_index);
	Nbin_vol_emis_aer =
	  max(Nbin_vol_emis_aer,
	      int(species_list_vol_emis_aer[j].second.size()));
      }
    Ns_vol_emis_aer = int(species_list_vol_emis_aer.size());
    // Additional variable: number of levels.
    if (this->option_process["with_volume_emission_aer"])
      {
	data_description_stream.SetSection("[volume_emission_aerosol]");
	data_description_stream.PeekValue("Nz", Nz_vol_emis_aer);
      }
    else
      Nz_vol_emis_aer = 0;
  }


  //! Checks that the configuration is acceptable.
  /*! In case any inconsistency is found, an exception is thrown.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::CheckConfiguration()
  {
    if (this->option_manage["snow_height"]
	&& this->input_files["meteo"]("SnowHeight").empty())
      throw "Snow height is needed but no input data file was provided.";

    if (this->scavenging_model == "microphysical-ph"
	&& !this->option_process["with_pH"])
      throw string("Scavenging model \"microphysical-pH\" cannot be used ")
	+ "without pH (option \"With_pH\").";

    if (this->scavenging_model == "microphysical"
	|| this->scavenging_model == "microphysical-ph")
      {
	// Checks if Henry constant is well initialized for each scavenged
	// species.
	for (int i = 0; i < this->Ns_scav; i++)
	  if (this->henry_constant.find(this->ScavengingName(i))
	      == this->henry_constant.end())
	    throw string("ERROR! Henry constant for species \"")
	      + this->ScavengingName(i)
	      + string("\" not found in section [henry].");
	
	// Checks if gas phase diffusivity constant is well initialized for
	// all scavenged species.
	for (int i = 0; i < this->Ns_scav; i++)
	  if (this->gas_phase_diffusivity.find(this->ScavengingName(i))
	      == this->gas_phase_diffusivity.end())
	    throw string("ERROR! Gas phase diffusivity constant")
	      + string(" for species \"") + this->ScavengingName(i)
	      + "\" not found in section [diffusivity].";
      }

    if (this->scavenging_model == "microphysical-ph")
      // Checks if dissolution heats are well initialized for all scavenged
      // species.
      for (int i = 0; i < this->Ns_scav; i++)
	if (dissolution_heat.find(this->ScavengingName(i))
	    == dissolution_heat.end())
	  throw string("ERROR! Dissolution heat for species \"")
	    + this->ScavengingName(i)
	    + string("\" not found in section [dissolution_heat].");

    if (!this->option_process["with_deposition_aer"]
	&& this->option_process["collect_dry_flux_aer"])
      throw string("Aerosol dry deposition fluxes cannot be collected") +
	" without deposition.";

    if (this->option_process["collect_wet_flux_aer"]
	&& (!this->option_process["with_scavenging_aer"]))
      throw string("Aerosol wet deposition fluxes cannot be collected") +
	" without scavenging.";
  }


  //! Checks whether an aerosol has initial conditions.
  /*!
    \param s species global index.
    \param b bin number.
    \return True if the aerosol has initial conditions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasInitialCondition_aer(int s, int b) const
  {
    return HasInitialCondition_aer(this->GetSpeciesName_aer(s), b);
  }


  //! Checks whether an aerosol has initial conditions.
  /*!
    \param name species name.
    \param b bin number.
    \return True if the aerosol has initial conditions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasInitialCondition_aer(string name, int b) const
  {
    unsigned int i;
    for (i = 0; i < species_list_ic_aer.size(); i++)
      if (species_list_ic_aer[i].first == name)
	return find(species_list_ic_aer[i].second.begin(),
		    species_list_ic_aer[i].second.end(), b)
	  != species_list_ic_aer[i].second.end();
    return false;
  }


  //! Checks whether an aerosol has boundary conditions.
  /*!
    \param s species global index.
    \param b bin number.
    \return True if the aerosol has boundary conditions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasBoundaryCondition_aer(int s, int b) const
  {
    return HasBoundaryCondition_aer(this->GetSpeciesName_aer(s), b);
  }


  //! Checks whether an aerosol has boundary conditions.
  /*!
    \param name species name.
    \param b bin number.
    \return True if the aerosol has boundary conditions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasBoundaryCondition_aer(string name, int b) const
  {
    unsigned int i;
    for (i = 0; i < species_list_bc_aer.size(); i++)
      if (species_list_bc_aer[i].first == name)
	return find(species_list_bc_aer[i].second.begin(),
		    species_list_bc_aer[i].second.end(), b)
	  != species_list_bc_aer[i].second.end();
    return false;
  }


  //! Returns the index in boundary conditions of a given aerosol.
  /*!
    \param s species global index.
    \param b bin number.
    \return The aerosol index in boundary conditions.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  vector<int>
  Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::BoundaryConditionIndex_aer(int s, int b) const
  {
    vector<int> output(2);
    unsigned int i;
    string name = this->GetSpeciesName_aer(s);
    for (i = 0; i < species_list_bc_aer.size(); i++)
      if (species_list_bc_aer[i].first == name)
	{
	  output[0] = int(i);
	  output[1] = find(species_list_bc_aer[i].second.begin(),
			   species_list_bc_aer[i].second.end(), b)
	    - species_list_bc_aer[i].second.begin();
	  return output;
	}
    throw string("Species \"") + name + string("_") + to_str(b)
      + "\" not found in boundary conditions.";
  }


  //! Checks whether aerosols in a given bin have deposition velocities.
  /*!
    \param b bin number.
    \return True if the aerosols in bin \a b has deposition velocities, false
    otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasDepositionVelocity_aer(int b) const
  {
    return find(bin_list_dep_aer.begin(), bin_list_dep_aer.end(), b)
      != bin_list_dep_aer.end();
  }


  //! Returns the index in deposition velocities of a given aerosol bin.
  /*!
    \param b bin number.
    \return The aerosol bin index in deposition velocities.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::DepositionVelocityIndex_aer(int b) const
  {
    return find(bin_list_dep_aer.begin(), bin_list_dep_aer.end(), b)
      - bin_list_dep_aer.begin();
  }


  //! Checks whether aerosols in a given bin have scavenging.
  /*!
    \param b bin number.
    \return True if the aerosols in bin \a b has deposition velocities, false
    otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasScavenging_aer(int b) const
  {
    return find(bin_list_scav_aer.begin(), bin_list_scav_aer.end(), b)
      != bin_list_scav_aer.end();
  }


  //! Returns the index in scavenging of a given aerosol bin.
  /*!
    \param b bin number.
    \return The aerosol bin index in scavenging.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::ScavengingIndex_aer(int b) const
  {
    return find(bin_list_scav_aer.begin(), bin_list_scav_aer.end(), b)
      - bin_list_scav_aer.begin();
  }


  //! Checks whether an aerosol has surface emissions.
  /*!
    \param s species global index.
    \param b bin number.
    \return True if the aerosol has surface emissions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasSurfaceEmission_aer(int s, int b) const
  {
    return HasSurfaceEmission_aer(this->GetSpeciesName_aer(s), b);
  }


  //! Checks whether an aerosol has surface emissions.
  /*!
    \param name species name.
    \param b bin number.
    \return True if the aerosol has surface emissions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasSurfaceEmission_aer(string name, int b) const
  {
    unsigned int i;
    for (i = 0; i < species_list_surf_emis_aer.size(); i++)
      if (species_list_surf_emis_aer[i].first == name)
	return find(species_list_surf_emis_aer[i].second.begin(),
		    species_list_surf_emis_aer[i].second.end(), b)
	  != species_list_surf_emis_aer[i].second.end();
    return false;
  }


  //! Returns the index in surface emissions of a given aerosol.
  /*!
    \param s species global index.
    \param b bin number.
    \return The aerosol index in surface emissions.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  vector<int>
  Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::SurfaceEmissionIndex_aer(int s, int b) const
  {
    vector<int> output(2);
    unsigned int i;
    string name = this->GetSpeciesName_aer(s);
    for (i = 0; i < species_list_surf_emis_aer.size(); i++)
      if (species_list_surf_emis_aer[i].first == name)
	{
	  output[0] = int(i);
	  output[1] = find(species_list_surf_emis_aer[i].second.begin(),
			   species_list_surf_emis_aer[i].second.end(), b)
	    - species_list_surf_emis_aer[i].second.begin();
	  return output;
	}
    throw string("Species \"") + name + string("_") + to_str(b)
      + "\" not found in surface emissions.";
  }


  //! Checks whether an aerosol has volume emissions.
  /*!
    \param s species global index.
    \param b bin number.
    \return True if the aerosol has volume emissions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasVolumeEmission_aer(int s, int b) const
  {
    return HasVolumeEmission_aer(this->GetSpeciesName_aer(s), b);
  }


  //! Checks whether an aerosol has volume emissions.
  /*!
    \param name species name.
    \param b bin number.
    \return True if the aerosol has volume emissions, false otherwise.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasVolumeEmission_aer(string name, int b) const
  {
    unsigned int i;
    for (i = 0; i < species_list_vol_emis_aer.size(); i++)
      if (species_list_vol_emis_aer[i].first == name)
	return find(species_list_vol_emis_aer[i].second.begin(),
		    species_list_vol_emis_aer[i].second.end(), b)
	  != species_list_vol_emis_aer[i].second.end();
   return false;
  }


  //! Returns the index in volume emissions of a given aerosol.
  /*!
    \param s aerosol global index.
    \return The aerosol index in volume emissions.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::VolumeEmissionIndex_aer(int s) const
  {
    string name = this->GetSpeciesName_aer(s);
    unsigned int i;
    for (i = 0; i < species_list_vol_emis_aer.size(); i++)
      if (species_list_vol_emis_aer[i].first == name)
	return int(i);
    throw string("Unable to find species \"") + name
      + "\" in aerosol volume emissions.";
  }


  //! Returns the name of an aerosol with volume emissions.
  /*!
    \param s species index in volume emissions (for aerosols).
    \return The species name.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  string Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::VolumeEmissionName_aer(int s) const
  {
    return species_list_vol_emis_aer.at(s).first;
  }


  //! Returns the global index of a species (aerosol) with volume emissions.
  /*!
    \param s species index in volume emissions (for aerosols).
    \return The species global index (aerosol).
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::VolumeEmissionGlobalIndex_aer(int s) const
  {
    return this->GetSpeciesIndex_aer(VolumeEmissionName_aer(s));
  }


  /////////////////////
  // INITIALIZATIONS //
  /////////////////////


  //! Allocates memory.
  /*! Allocates grids and fields.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Allocate()
  {
    Polair3DChemistry<T, ClassAdvection, ClassDiffusion,
      ClassChemistry>::Allocate();

    /*** Additional grids ***/

    GridZ5D = RegularGrid<T>(this->Nz);
    GridY5D = RegularGrid<T>(this->y_min, this->Delta_y, this->Ny);
    GridX5D = RegularGrid<T>(this->x_min, this->Delta_x, this->Nx);

    GridY5D_interf_bc = RegularGrid<T>(this->y_min - this->Delta_y,
				       T(this->Ny + 1) * this->Delta_y, 2);
    GridX5D_interf_bc = RegularGrid<T>(this->x_min - this->Delta_x,
				       T(this->Nx + 1) * this->Delta_x, 2);

    BinBound_aer.resize(this->Nbin_aer + 1);
    // Reads bin bounds in micrometers and converts it to meters.
    for (int i = 0; i < this->Nbin_aer + 1; i++)
      BinBound_aer(i) = 1.e-6 * convert<T>(bin_list[i]);
    GridB4D_aer = RegularGrid<T>(this->Nbin_aer);
    GridB5D_aer = RegularGrid<T>(this->Nbin_aer);

    GridS4D_aer = RegularGrid<T>(this->Ns_aer);
    GridS5D_aer = RegularGrid<T>(this->Ns_aer);

    GridS_bc_aer = RegularGrid<T>(Ns_bc_aer);
    GridB_bc_aer = RegularGrid<T>(Nbin_bc_aer);

    GridB_dep_aer = RegularGrid<T>(Nbin_dep_aer);
    GridB_scav_aer = RegularGrid<T>(Nbin_scav_aer);

    GridS_surf_emis_aer = RegularGrid<T>(Ns_surf_emis_aer);
    GridB_surf_emis_aer = RegularGrid<T>(Nbin_surf_emis_aer);

    GridS_vol_emis_aer = RegularGrid<T>(Ns_vol_emis_aer);
    GridB_vol_emis_aer = RegularGrid<T>(Nbin_vol_emis_aer);
    GridZ_vol_emis_aer = RegularGrid<T>(Nz_vol_emis_aer);

    GridZ5D.SetVariable(2);
    GridZ5D.SetDuplicate(false);

    GridY5D.SetVariable(3);
    GridY5D.SetDuplicate(false);
    GridY5D_interf_bc.SetVariable(3);
    GridY5D_interf_bc.SetDuplicate(false);

    GridX5D.SetVariable(4);
    GridX5D.SetDuplicate(false);
    GridX5D_interf_bc.SetVariable(4);
    GridX5D_interf_bc.SetDuplicate(false);

    /*** Meteorological fields ***/

    SurfaceTemperature_i.Resize(this->GridY2D, this->GridX2D);
    SurfaceTemperature_f.Resize(this->GridY2D, this->GridX2D);
    FileSurfaceTemperature_i.Resize(this->GridY2D, this->GridX2D);
    FileSurfaceTemperature_f.Resize(this->GridY2D, this->GridX2D);

    SurfacePressure_i.Resize(this->GridY2D, this->GridX2D);
    SurfacePressure_f.Resize(this->GridY2D, this->GridX2D);
    FileSurfacePressure_i.Resize(this->GridY2D, this->GridX2D);
    FileSurfacePressure_f.Resize(this->GridY2D, this->GridX2D);

    LiquidWaterContent_i.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);
    FileLiquidWaterContent_i.Resize(this->GridZ3D,
				    this->GridY3D, this->GridX3D);
    FileLiquidWaterContent_f.Resize(this->GridZ3D,
				    this->GridY3D, this->GridX3D);

    if (this->option_process["compute_deposition_aer"])
      {
	FirstLevelWindModule_i.Resize(this->GridY2D, this->GridX2D);
	FirstLevelWindModule_f.Resize(this->GridY2D, this->GridX2D);

	SnowHeight_i.Resize(this->GridY2D, this->GridX2D);
	SnowHeight_f.Resize(this->GridY2D, this->GridX2D);
	FileSnowHeight_i.Resize(this->GridY2D, this->GridX2D);
	FileSnowHeight_f.Resize(this->GridY2D, this->GridX2D);
      }

    RelativeHumidity_i.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);
    RelativeHumidity_f.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);

    /*** Boundary conditions ***/

    // Along Z.
    BoundaryCondition_z_aer_i.Resize(GridS_bc_aer, GridB_bc_aer,
				     this->GridY4D, this->GridX4D);
    FileBoundaryCondition_z_aer_i.Resize(GridS_bc_aer, GridB_bc_aer,
					 this->GridY4D, this->GridX4D);
    FileBoundaryCondition_z_aer_f.Resize(GridS_bc_aer, GridB_bc_aer,
					 this->GridY4D, this->GridX4D);

    // Along Y.
    BoundaryCondition_y_aer_i.Resize(GridS_bc_aer, GridB_bc_aer,
				     this->GridZ5D, this->GridY5D_interf_bc,
				     this->GridX5D);
    FileBoundaryCondition_y_aer_i.Resize(GridS_bc_aer, GridB_bc_aer,
					 this->GridZ5D,
					 this->GridY5D_interf_bc,
					 this->GridX5D);
    FileBoundaryCondition_y_aer_f.Resize(GridS_bc_aer, GridB_bc_aer,
					 this->GridZ5D,
					 this->GridY5D_interf_bc,
					 this->GridX5D);

    // Along X.
    BoundaryCondition_x_aer_i.Resize(GridS_bc_aer, GridB_bc_aer,
				     this->GridZ5D, this->GridY5D,
				     this->GridX5D_interf_bc);
    FileBoundaryCondition_x_aer_i.Resize(GridS_bc_aer, GridB_bc_aer,
					 this->GridZ5D,
					 this->GridY5D,
					 this->GridX5D_interf_bc);
    FileBoundaryCondition_x_aer_f.Resize(GridS_bc_aer, GridB_bc_aer,
					 this->GridZ5D,
					 this->GridY5D,
					 this->GridX5D_interf_bc);

    /*** Loss terms ***/

    DepositionVelocity_aer_i.Resize(GridB_dep_aer,
				    this->GridY3D, this->GridX3D);
    DepositionVelocity_aer_f.Resize(GridB_dep_aer,
				    this->GridY3D, this->GridX3D);
    FileDepositionVelocity_aer_i.Resize(GridB_dep_aer,
					this->GridY3D, this->GridX3D);
    FileDepositionVelocity_aer_f.Resize(GridB_dep_aer,
					this->GridY3D, this->GridX3D);
    
    ScavengingCoefficient_aer_i.Resize(GridB_scav_aer, this->GridZ4D,
				       this->GridY4D, this->GridX4D);
    ScavengingCoefficient_aer_f.Resize(GridB_scav_aer, this->GridZ4D,
				       this->GridY4D, this->GridX4D);

    /*** Source terms ***/

    // Surface emissions.
    SurfaceEmission_aer_i.Resize(GridS_surf_emis_aer, GridB_surf_emis_aer,
				 this->GridY4D, this->GridX4D);
    SurfaceEmission_aer_f.Resize(GridS_surf_emis_aer, GridB_surf_emis_aer,
				 this->GridY4D, this->GridX4D);
    FileSurfaceEmission_aer_i.Resize(GridS_surf_emis_aer, GridB_surf_emis_aer,
				     this->GridY4D, this->GridX4D);
    FileSurfaceEmission_aer_f.Resize(GridS_surf_emis_aer, GridB_surf_emis_aer,
				     this->GridY4D, this->GridX4D);

    // Volume emissions.
    VolumeEmission_aer_i.Resize(GridS_vol_emis_aer, GridB_vol_emis_aer,
				this->GridZ_vol_emis_aer, this->GridY5D,
				this->GridX5D);
    VolumeEmission_aer_f.Resize(GridS_vol_emis_aer, GridB_vol_emis_aer,
				this->GridZ_vol_emis_aer, this->GridY5D,
				this->GridX5D);

    FileVolumeEmission_aer_i.Resize(GridS_vol_emis_aer, GridB_vol_emis_aer,
				    this->GridZ_vol_emis_aer, this->GridY5D,
				    this->GridX5D);
    FileVolumeEmission_aer_f.Resize(GridS_vol_emis_aer, GridB_vol_emis_aer,
				    this->GridZ_vol_emis_aer, this->GridY5D,
				    this->GridX5D);

    /*** Sources (source splitting) ***/

    if (this->source_splitting && this->option_process["with_chemistry"])
      {
	Source_aer_i.Resize(GridS5D_aer, GridB5D_aer,
			    this->GridZ5D, this->GridY5D, this->GridX5D);
	Source_aer_f.Resize(GridS5D_aer, GridB5D_aer,
			    this->GridZ5D, this->GridY5D, this->GridX5D);
      }
    
    /*** State ***/

    this->Concentration_aer.Resize(GridS5D_aer, GridB5D_aer,
				   GridZ5D, GridY5D, GridX5D);

    /*** Deposition fluxes ***/

    if (this->option_process["collect_dry_flux_aer"])
      DryDepositionFlux_aer.Resize(GridS4D_aer, GridB_dep_aer, this->GridY4D,
				   this->GridX4D);

    if (this->option_process["collect_wet_flux_aer"])
      {
	WetDepositionFlux_aer.Resize(GridS4D_aer, GridB_scav_aer,
				     this->GridY4D, this->GridX4D);
	InCloudWetDepositionFlux_aer.Resize(GridS4D_aer, GridB_scav_aer,
					    this->GridY4D, this->GridX4D);
      }

    /*** Aerosol physical properties ***/

    // Density per bin in kg / m^3.
    Density_aer.resize(this->Nbin_aer);

    WetDiameter_aer.Resize(GridB4D_aer, this->GridZ4D, this->GridY4D,
			   this->GridX4D);

    pH.Resize(this->GridZ3D, this->GridY3D, this->GridX3D);
    // Default cloud pH.
    pH.Fill(4.5);
  }


  //! Model initialization.
  /*! It reads the configuration, allocates memory and reads the values of the
    fields at the beginning of the simulation.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Init()
  {
    Polair3DTransport<T, ClassAdvection, ClassDiffusion>::Init();

    Polair3DChemistry<T, ClassAdvection, ClassDiffusion, ClassChemistry>
      ::Allocate();

    Polair3DChemistry<T, ClassAdvection, ClassDiffusion, ClassChemistry>
      ::InitAllData();

    Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
      ::Allocate();

    /*** Aerosol density ***/

    // Density per bin in kg / m^3.
    Density_aer = fixed_density_aer;

    /*** Land data ***/

    if (this->option_process["compute_deposition_aer"])
      {
	string land_data_file, LUC_file;
	this->config.SetSection("[data]");
	this->config.PeekValue("Land_data", land_data_file);
	ConfigStream land_data_stream(land_data_file);
	// Land use cover.
	land_data_stream.SetSection("[LUC]");
	land_data_stream.PeekValue("Nland", Nland);
	LUC.Resize(Nland, this->Ny, this->Nx);
	land_data_stream.PeekValue("LUC", LUC_file);
	FormatBinary<float>().Read(LUC_file, LUC);
	// Land data.
	// Reads roughness heights for all land categories and all five
	// seasons.
	land_data_stream.SetSection("[roughness]");
	LandRoughnessHeight.resize(Nland, 5);
	for (int c = 0; c < Nland; c++)
	  for (int s = 0; s < 5; s++)
	    land_data_stream.GetNumber(LandRoughnessHeight(c, s));
	// Reads characteristic radius of small receptors for all land
	// categories and all five seasons.
	land_data_stream.SetSection("[small_radius]");
	SmallRadius.resize(Nland, 5);
	for (int c = 0; c < Nland; c++)
	  for (int s = 0; s < 5; s++)
	    land_data_stream.GetNumber(SmallRadius(c, s));
	// Reads characteristic radius of large receptors for all land
	// categories and all five seasons.
	land_data_stream.SetSection("[large_radius]");
	LargeRadius.resize(Nland, 5);
	for (int c = 0; c < Nland; c++)
	  for (int s = 0; s < 5; s++)
	    land_data_stream.GetNumber(LargeRadius(c, s));
	// Reads Zhang coefficients alpha and gamma.
	land_data_stream.SetSection("[zhang_alpha]");
	ZhangAlpha.resize(Nland);
	for (int c = 0; c < Nland; c++)
	  land_data_stream.GetNumber(ZhangAlpha(c));
	land_data_stream.SetSection("[zhang_gamma]");
	ZhangGamma.resize(Nland);
	for (int c = 0; c < Nland; c++)
	  land_data_stream.GetNumber(ZhangGamma(c));
      }

    /*** Input fields ***/

    Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
      ::InitAllData();

    /*** Initial conditions for aerosols ***/

    if (this->option_manage["initial_condition_aer"])
      {
    	string species, species_bin;
    	vector<int> bin;

    	this->Concentration_aer.SetZero();
	
    	for (int i = 0; i < Ns_ic_aer; i++)
    	  {
    	    species = species_list_ic_aer[i].first;
    	    bin = species_list_ic_aer[i].second;
    	    for (int j = 0; j < int(bin.size()); j++)
    	      {
    		species_bin = species + string("_") + to_str(bin[j]);
    		string filename
    		  = this->input_files["initial_condition_aer"](species_bin);
    		int index = this->GetSpeciesIndex_aer(species);
    		Data<T, 3>
    		  Concentration_tmp(&this->Concentration_aer(index, bin[j],
							     0, 0, 0),
    				    shape(this->Nz, this->Ny, this->Nx));
		if (is_num(filename))
		  Concentration_tmp.Fill(to_num<T>(filename));
		else
		  FormatBinary<float>().Read(filename, Concentration_tmp);
    	      }
    	  }
      }

    /*** Chemical mechanism ***/

    this->Chemistry_.Init(*this);

    /*** Deposition ***/

    if (this->option_process["collect_dry_flux_aer"])
      for (int s = 0; s < this->Ns_aer; s++)
	for (int b = 0; b < Nbin_dep_aer; b++)
	  for (int j = 0; j < this->Ny; j++)
	    for (int i = 0; i < this->Nx; i++)
	      DryDepositionFlux_aer(s, b, j, i) =
		DepositionVelocity_aer_f(b, j, i)
		* this->Concentration_aer(s, bin_list_dep_aer[b], 0, j, i);

    if (this->option_process["collect_wet_flux_aer"])
      {
	this->WetDepositionFlux_aer.SetZero();
	this->InCloudWetDepositionFlux_aer.SetZero();
	for (int s = 0; s < this->Ns_aer; s++)
	  for (int b = 0; b < Nbin_scav_aer; b++)
	    for (int j = 0; j < this->Ny; j++)
	      for (int i = 0; i < this->Nx; i++)
		if (this->Rain_i(j, i) > 0.)
		  for (int k = 0; k < this->Nz; k++)
		    if (this->CloudHeight_f(j, i) < this->GridZ4D(k))
		      WetDepositionFlux_aer(s, b, j, i) =
			this->Concentration_aer(s, bin_list_scav_aer[b],
						k, j, i)
			/ this->Delta_t
			* (1.- exp(- this->Delta_t
				   * ScavengingCoefficient_aer_f(b, k, j, i)))
			* (this->GridZ4D_interf(k+1)
			   - this->GridZ4D_interf(k));
      }
  }


  //! Model initialization for each step.
  /*! It reads on file the data needed for the current step.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::InitStep()
  {
    int k, i, j;
    string species, species_bin;
    vector<int> bin;
    string filename;
    Date date;
    T Delta_t;

    Polair3DChemistry<T, ClassAdvection, ClassDiffusion,
      ClassChemistry>::InitStep();

    /*** Surface temperature ***/
    
    if (this->option_manage["surface_temperature"])
      this->UpdateData("meteo", "SurfaceTemperature",
		       FileSurfaceTemperature_i, FileSurfaceTemperature_f,
		       SurfaceTemperature_i);

    /*** Surface pressure ***/
    
    if (this->option_manage["surface_pressure"])
      this->UpdateData("meteo", "SurfacePressure", FileSurfacePressure_i,
		       FileSurfacePressure_f, SurfacePressure_i);

    /*** First level wind module ***/

    FirstLevelWindModule_i.GetArray() = FirstLevelWindModule_f.GetArray();
    T zonal_wind, meridional_wind;
    if (this->option_manage["first_level_wind_module"])
      for (j = 0; j < this->Ny; j++)
	for (i = 0; i < this->Nx; i++)
	  {
	    zonal_wind = 0.5 * (this->ZonalWind_i(0, j, i)
				+ this->ZonalWind_i(0, j, i + 1));
	    meridional_wind = 0.5 * (this->MeridionalWind_i(0, j, i)
				     + this->MeridionalWind_i(0, j + 1, i));
	    FirstLevelWindModule_f(j, i) =
	      sqrt(zonal_wind * zonal_wind
		   + meridional_wind * meridional_wind);
	  }

    /*** Liquid water content ***/
    
    if (this->option_manage["liquid_water_content"])
      this->UpdateData("meteo", "LiquidWaterContent",
		       FileLiquidWaterContent_i, FileLiquidWaterContent_f,
		       LiquidWaterContent_i);

    /*** Relative humidity ***/
    
    RelativeHumidity_i.GetArray() = RelativeHumidity_f.GetArray();
    for (k = 0; k < this->Nz; k++)
      for (j = 0; j < this->Ny; j++)
	for (i = 0; i < this->Nx; i++)
	  {
	    _compute_relative_humidity(&(this->SpecificHumidity_f(k, j, i)),
				       &(this->Temperature_f(k, j, i)),
				       &(this->Pressure_f(k, j, i)),
				       &RelativeHumidity_f(k, j, i));
	    RelativeHumidity_f(k, j, i) =
	      min(max(RelativeHumidity_f(k, j, i), 0.), 1.);
	  }

    /*** Boundary conditions ***/

    if (this->option_manage["boundary_condition_aer"])
      for (i = 0; i < Ns_bc_aer; i++)
	{
	  species = species_list_bc_aer[i].first;
	  bin = species_list_bc_aer[i].second;
	  for (j = 0; j < int(bin.size()); j++)
	    {
	      species_bin = species + string("_") + to_str(bin[j]);
	      string filename
		= this->input_files["boundary_condition_aer"](species_bin);
	      date = this->input_files["boundary_condition_aer"].GetDateMin();
	      Delta_t
		= this->input_files["boundary_condition_aer"].GetDelta_t();
	      this->UpdateData(find_replace(filename, "&c", "z"), date,
			       Delta_t, FileBoundaryCondition_z_aer_i,
			       FileBoundaryCondition_z_aer_f, i, j,
			       BoundaryCondition_z_aer_i);
	      this->UpdateData(find_replace(filename, "&c", "y"), date,
			       Delta_t, FileBoundaryCondition_y_aer_i,
			       FileBoundaryCondition_y_aer_f, i, j,
			       BoundaryCondition_y_aer_i);
	      this->UpdateData(find_replace(filename, "&c", "x"), date,
			       Delta_t, FileBoundaryCondition_x_aer_i,
			       FileBoundaryCondition_x_aer_f, i, j,
			       BoundaryCondition_x_aer_i);
	    }
	}

    /*** Wet aerosol diameter ***/

    this->InitWetDiameter_aer(RelativeHumidity_i, this->Temperature_i,
			      WetDiameter_aer);

    /*** Deposition velocities ***/

    if (this->option_manage["deposition_velocity_aer"])
      if (!this->option_process["compute_deposition_aer"])
	for (i = 0; i < Nbin_dep_aer; i++)
	  this->UpdateData("deposition_velocity_aer",
			   to_str(bin_list_dep_aer[i]),
			   FileDepositionVelocity_aer_i,
			   FileDepositionVelocity_aer_f,
			   i, DepositionVelocity_aer_i,
			   DepositionVelocity_aer_f);
      else
	{
	  DepositionVelocity_aer_i.GetArray()
	    = DepositionVelocity_aer_f.GetArray();
	  this->InitDepositionVelocity(this->Temperature_f, this->Pressure_f,
				       this->SurfaceTemperature_f,
				       this->SurfacePressure_f,
				       FirstLevelWindModule_f,
				       RelativeHumidity_f, SnowHeight_f,
				       DepositionVelocity_aer_f);
	}

    /*** Scavenging coefficients ***/

    if (this->option_manage["scavenging_coefficient"]
	&& this->scavenging_model == "microphysical-ph")
      InitScavengingCoefficient(this->Temperature_f, this->Pressure_f,
				LiquidWaterContent_i, pH,
				this->CloudHeight_f, this->Rain_i,
				this->ScavengingCoefficient_f);

    if (this->option_manage["scavenging_coefficient_aer"]
	&& this->option_process["with_scavenging_aer"])
	{
	  ScavengingCoefficient_aer_i.GetArray()
	    = ScavengingCoefficient_aer_f.GetArray();
	  InitScavengingCoefficient_aer(this->Temperature_f,
					this->Pressure_f,
					WetDiameter_aer,
					this->CloudHeight_f,
					this->Rain_i,
					ScavengingCoefficient_aer_f);
	}

    /*** Surface emissions ***/

    if (this->option_manage["surface_emission_aer"])
      for (i = 0; i < Ns_surf_emis_aer; i++)
	{
	  species = species_list_surf_emis_aer[i].first;
	  bin = species_list_surf_emis_aer[i].second;
	  for (j = 0; j < int(bin.size()); j++)
	    {
	      species_bin = species + string("_") + to_str(bin[j]);
	      this->UpdateData("surface_emission_aer", species_bin,
			       FileSurfaceEmission_aer_i,
			       FileSurfaceEmission_aer_f, i, j,
			       SurfaceEmission_aer_i,
			       SurfaceEmission_aer_f);
	    }
	}

    /*** Volume emissions ***/

    if (this->option_manage["volume_emission_aer"])
      for (i = 0; i < Ns_vol_emis_aer; i++)
	{
	  species = species_list_vol_emis_aer[i].first;
	  bin = species_list_vol_emis_aer[i].second;
	  for (j = 0; j < int(bin.size()); j++)
	    {
	      species_bin = species + string("_") + to_str(bin[j]);
	      this->UpdateData("volume_emission_aer", species_bin,
			       FileVolumeEmission_aer_i,
			       FileVolumeEmission_aer_f, i, j,
			       VolumeEmission_aer_i,
			       VolumeEmission_aer_f);
	    }
	}
  }


  //! Initializes wet aerosol diameters.
  /*! Computes wet aerosol diameters for aerosols using Gerber formula.
    \param RelativeHumidity_ relative humidity.
    \param Temperature_ temperature (K).
    \param WetDiameter_aer_ (output) wet aerosol diameter (m).
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::InitWetDiameter_aer(Data<T, 3>& RelativeHumidity_,
			Data<T, 3>& Temperature_,
			Data<T, 4>& WetDiameter_aer_)
  {
    int i, j, k, b;

    Array<T, 1> MeanDiameter(this->Nbin_aer);
    for (b = 0; b < this->Nbin_aer; b++)
      MeanDiameter(b) = 1.e6 * sqrt(BinBound_aer(b + 1) * BinBound_aer(b));

    for (b = 0; b < this->Nbin_aer; b++)
      for (k = 0; k < this->Nz; k++)
	for (j = 0; j < this->Ny; j++)
	  for (i = 0; i < this->Nx; i++)
	    _gerber_wet_diameter(&RelativeHumidity_(k, j, i),
				 &Temperature_(k, j, i), &MeanDiameter(b),
				 &WetDiameter_aer_(b, k, j, i));
    // Back to meters.
    WetDiameter_aer_() *= 1.e-6;
  }


  //! Initializes deposition velocities for aerosols.
  /*! Computes deposition velocities for aerosols on the basis of Zhang
    parameterization.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::InitDepositionVelocity(Data<T, 3>& Temperature_,
			   Data<T, 3>& Pressure_,
			   Data<T, 2>& SurfaceTemperature_,
			   Data<T, 2>& SurfacePressure_,
			   Data<T, 2>& FirstLevelWindModule_,
			   Data<T, 3>& RelativeHumidity_,
			   Data<T, 2>& SnowHeight_,
			   Data<T, 3>& DepositionVelocity_aer_)
  {
    int i, j, c, b;
    T seconds(this->current_date.GetNumberOfSeconds());
    Array<T, 1> MeanDiameter(this->Nbin_dep_aer);
    for (i = 0; i < this->Nbin_dep_aer; i++)
      MeanDiameter(i) = sqrt(BinBound_aer(bin_list_dep_aer[i] + 1)
			     * BinBound_aer(bin_list_dep_aer[i]));

    // 1D variables.
    Array<T, 1> LUC_1D(Nland);
    Array<T, 1> DepositionVelocity_1D(this->Nbin_dep_aer);

    DepositionVelocity_aer_.SetZero();
    for (j = 0; j < this->Ny; j++)
      for (i = 0; i < this->Nx; i++)
	{
	  for (c = 0; c < Nland; c++)
	    LUC_1D(c) = LUC(c, j, i);
	  _compute_dry_deposition(&seconds, &this->Nbin_dep_aer, &Nland,
				  MeanDiameter.data(), Density_aer.data(),
				  LUC_1D.data(), &RelativeHumidity_(0, j, i),
				  &(this->GridZ3D(0)),
				  &FirstLevelWindModule_(j, i),
				  &Temperature_(0, j, i),
				  &SurfaceTemperature_(j, i),
				  &Pressure_(0, j, i),
				  &SurfacePressure_(j, i), &SnowHeight_(j, i),
				  LandRoughnessHeight.data(),
				  ZhangGamma.data(), ZhangAlpha.data(),
				  SmallRadius.data(), LargeRadius.data(),
				  DepositionVelocity_1D.data());
	  for (b = 0; b < this->Nbin_dep_aer; b++)
	    DepositionVelocity_aer_(bin_list_dep_aer[b], j, i) =
	      DepositionVelocity_1D(b);
	}
  }

 //! Initializes scavenging coefficients.
  /*! Computes below-cloud scavenging coefficients of a reversibly soluble
    gas.
    \param Temperature_ temperature (K).
    \param Pressure_ pressure (Pa).
    \param LiquidWaterContent_ liquid water content (kg/kg).
    \param pH_ pH of cloud droplet.
    \param CloudHeight_ cloud basis height (m).
    \param Rain_ rain intensity (mm/h).
    \param ScavengingCoefficient_ (output) the scavenging coefficients
    (s^{-1}).
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::InitScavengingCoefficient(Data<T, 3>& Temperature_,
			      Data<T, 3>& Pressure_,
			      Data<T, 3>& LiquidWaterContent_,
			      Data<T, 3>& pH_,
			      Data<T, 2>& CloudHeight_,
			      Data<T, 2>& Rain_,
			      Data<T, 4>& ScavengingCoefficient_)
  {
    int i, j, k, s;

    int Ns_scav = ScavengingCoefficient_.GetLength(0);
    int Nz = ScavengingCoefficient_.GetLength(1);
    int Ny = ScavengingCoefficient_.GetLength(2);
    int Nx = ScavengingCoefficient_.GetLength(3);

    T concH, pH_loc(0.), lwc_loc, factor;
    T H1, H2, H3, H4, H5, H6, Kw;
    T Pr = 8.2e-2;

    bool has_scavenging_SO2 = this->HasScavenging("SO2");
    bool has_scavenging_NH3 = this->HasScavenging("NH3");
    bool has_scavenging_HNO3 = this->HasScavenging("HNO3");
    bool has_scavenging_HONO = this->HasScavenging("HONO");
    bool has_scavenging_HNO2 = this->HasScavenging("HNO2");
    bool has_scavenging_HCL = this->HasScavenging("HCL");

    int index_SO2 = this->ScavengingIndex("SO2");
    int index_NH3 = this->ScavengingIndex("NH3");
    int index_HNO3 = this->ScavengingIndex("HNO3");
    int index_HONO = this->ScavengingIndex("HONO");
    int index_HNO2 = this->ScavengingIndex("HNO2");
    int index_HCL = this->ScavengingIndex("HCL");

    // Henry constants of scavenged species.
    Array<T, 4> henry_constant_(Ns_scav, Nz, Ny, Nx);
    
    // Heat of dissolution.
    Array<T, 1> dissolution_heat_(Ns_scav);
    // Gas phase diffusivities of scavenged species.
    Array<T, 1> gas_phase_diffusivity_(Ns_scav);

    // Extracts species data for scavenged species only.
    for (s = 0; s < this->Ns_scav; s++)
      {
	dissolution_heat_(s) = dissolution_heat[this->ScavengingName(s)];
	gas_phase_diffusivity_(s) =
	  this->gas_phase_diffusivity[this->ScavengingName(s)];
      }

    // Extracts species data for scavenged species only.
    for (s = 0; s < Ns_scav; s++)
      for (k = 0; k < Nz; k++)
	for (j = 0; j < Ny; j++)
	  for (i = 0; i < Nx; i++)
	    henry_constant_(s, k, j, i) =
	      this->henry_constant[this->ScavengingName(s)];
 
    // Computes Henry constants at current temperature.
    for (k = 0; k < Nz; k++)
      for (j = 0; j < Ny; j++)
	for (i = 0; i < Nx; i++)
	  {
	    factor = 1. / 298. - 1. / Temperature_(k, j, i);
	    factor /= 0.001985;
	    for (s = 0; s < this->Ns_scav; s++)
	      henry_constant_(s, k, j, i)
		*= exp(dissolution_heat_(s) * factor);
	  }

    // Modifies Henry constants according to pH.
    for (j = 0; j < Ny; j++)
      for (i = 0; i < Nx; i++)
	{
	  for (k = 1; k < Nz; k++)
	    {
	      // Converts from kg / kg to g / m^3.
	      lwc_loc = LiquidWaterContent_(k, j, i) * 1000.0
		* Pressure_(k, j, i) / 101325.0 * 28.97 / Pr
		/ Temperature_(k, j, i);
	      
	      // Default cloud pH.
	      pH_loc = 4.5;
		
	      if (CloudHeight_(j, i) >= this->GridZ4D(k-1) &&
		  CloudHeight_(j, i) < this->GridZ4D(k) &&
		  lwc_loc >= lwc_cloud_threshold
		  && pH_(k, j, i) > 0.)
		// Assumed bounds on H+ concentration implies that pH is
		// between 4 and 6.
		pH_loc = min(max(pH_(k, j, i), 4.), 6.);
	    }

	  // H+ concentration.
	  concH = pow(10., -pH_loc);

	  for (k = 0; k < Nz; k++)
	    {
	      factor = 1. / 298. - 1. / Temperature_(k, j, i);
	      H1 = 0.0130 * exp(-2000. * factor);
	      H2 = 6.3e-8 * exp(-1495. * factor);
	      H3 = 1.7e-5 * exp(4325. * factor);
	      H4 = 15.1 * exp(-8700. * factor);
	      H5 = 5.9e-4 * exp(1760. * factor);
	      H6 = 1.7e6 * exp(-6900. * factor);
	      Kw = 1.e-14 * exp(6700. * factor);
		
	      if (has_scavenging_SO2)
		henry_constant_(index_SO2, k, j, i) *=
		  1. + H1 / concH + H1 * H2 / (concH*concH);
	      if (has_scavenging_NH3)
		henry_constant_(index_NH3, k, j, i) *= 1. + H3 * concH / Kw;
	      if (has_scavenging_HNO3)
		henry_constant_(index_HNO3, k, j, i) *= 1. + H4 / concH;
	      if (has_scavenging_HONO)
		henry_constant_(index_HONO, k, j, i) *= 1. + H5 / concH;
	      if (has_scavenging_HNO2)
		henry_constant_(index_HNO2, k, j, i) *= 1. + H5 / concH;
	      if (has_scavenging_HCL)
		henry_constant_(index_HCL, k, j, i) *= 1. + H6 / concH;
	    }
	}

    _compute_scavenging_coefficient(&Nx, &Ny, &Nz, &Ns_scav,
				    &Nx, &Ny, &Nz,
				    this->GridZ4D.GetArray().data(),
				    gas_phase_diffusivity_.data(),
				    henry_constant_.data(),
				    Temperature_.GetData(),
				    Pressure_.GetData(),
				    Rain_.GetData(),
				    CloudHeight_.GetData(),
				    ScavengingCoefficient_.GetData());
  }


  //! Initializes scavenging coefficients for aerosols.
  /*! Computes below-cloud scavenging coefficients for aerosols.
    \param Temperature_ temperature (K).
    \param Pressure_ pressure (Pa).
    \param WetDiameter_aer_ (output) wet aerosol diameter (m).
    \param CloudHeight_ cloud basis height (m).
    \param Rain_ rain intensity (mm/h).
    \param ScavengingCoefficient_aer_ (output) the scavenging coefficients
    (s^{-1}).
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::InitScavengingCoefficient_aer(Data<T, 3>& Temperature_,
				  Data<T, 3>& Pressure_,
				  Data<T, 4>& WetDiameter_aer_,
				  Data<T, 2>& CloudHeight_,
				  Data<T, 2>& Rain_,
				  Data<T, 4>& ScavengingCoefficient_aer_)
  {
    int i, j, k, b;

    // 1D variables.
    Array<T, 1> Coefficient(this->Nbin_scav_aer);
    Array<T, 1> wet_diameter_loc(this->Nbin_scav_aer);

    ScavengingCoefficient_aer_.SetZero();
    for (k = 0; k < this->Nz; k++)
      for (j = 0; j < this->Ny; j++)
	for (i = 0; i < this->Nx; i++)
	  {
	    for (b = 0; b < this->Nbin_scav_aer; b++)
	      wet_diameter_loc(b) =
		WetDiameter_aer_(bin_list_scav_aer[b], k, j, i);

	    _compute_scavenging_coefficient_aer(&this->Nbin_scav_aer,
						&Temperature_(k, j, i),
						&Pressure_(k, j, i),
						&Rain_(j, i),
						wet_diameter_loc.data(),
						Density_aer.data(),
						&(this->GridZ3D(k)),
						&CloudHeight_(j, i),
						Coefficient.data());
	    for (b = 0; b < this->Nbin_scav_aer; b++)
	      ScavengingCoefficient_aer_(bin_list_scav_aer[b], k, j, i) =
		Coefficient(b);
	  }
  }


  /////////////////
  // INTEGRATION //
  /////////////////


  //! Performs one advection step.
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Advection()
  {
    this->Advection_.Forward(*this);
    this->Advection_.Forward_aer(*this);
  }
  
  
  //! Performs one diffusion step.
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Diffusion()
  {
    this->Diffusion_.Forward(*this);
    this->Diffusion_.Forward_aer(*this);
  }


  //! Performs the time integration of aerosol point emissions.
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::PointEmission_aer()
  {
    int e, i, j, k, s, b;
    int Nsource(point_emission_list_aer.size());
    T quantity;
    // Loop over all point sources.
    for (e = 0; e < Nsource; e++)
      {
	i = convert<int>(point_emission_list_aer[e]["i"]);
	j = convert<int>(point_emission_list_aer[e]["j"]);
	k = convert<int>(point_emission_list_aer[e]["k"]);
	s = this->GetSpeciesIndex_aer(point_emission_list_aer[e]["species"]);
	b = convert<int>(point_emission_list_aer[e]["bin"]);
	if (point_emission_list_aer[e]["type"] == "puff")
	  {
	    Date date(point_emission_list_aer[e]["date"]);
	    if (this->current_date <= date && date < this->next_date)
	      this->Concentration_aer(s, b, k, j, i) +=
		convert<T>(point_emission_list_aer[e]["quantity"])
		/ (this->CellWidth_x(i) * this->CellWidth_y(j)
		   * this->CellWidth_z(k));
	  }
	else if (point_emission_list_aer[e]["type"] == "continuous")
	  {
	    Date date_beg(point_emission_list_aer[e]["date_beg"]);
	    Date date_end(point_emission_list_aer[e]["date_end"]);
	    if (this->next_date > date_beg && this->current_date < date_end)
	      {
		quantity = this->Delta_t
		  - max(0., date_beg.GetSecondsFrom(this->current_date))
		  - max(0., this->next_date.GetSecondsFrom(date_end));
		quantity *= convert<T>(point_emission_list_aer[e]["rate"]);
		this->Concentration_aer(s, b, k, j, i) += quantity
		  / (this->CellWidth_x(i) * this->CellWidth_y(j)
		     * this->CellWidth_z(k));
	      }
	  }
      }
  }


  //! Performs one chemistry step.
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Chemistry()
  {
    this->Chemistry_.Forward(*this);
    this->Chemistry_.Forward_aer(*this);
  }


  //! Performs one step forward.
  /*! It performs one advection step, then one diffusion step and finally
    integrates chemistry. The first two steps are split (operator
    splitting). The last step (chemistry) may be split or partially coupled
    through source splitting.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::Forward()
  {
    int i, j, k;

    if (this->source_splitting && this->option_process["with_chemistry"])
      {
	this->Source_f.GetArray() = this->Concentration.GetArray();
	this->Source_aer_f.GetArray() = this->Concentration_aer.GetArray();
      }

    if (this->option_process["with_advection"])
      this->Advection();
    //! Dry fluxes are collected before dry deposition is performed
    //! in diffusion.
    if (this->option_process["collect_dry_flux"])
      for (int s = 0; s < this->Ns_dep; s++)
	for (int j = 0; j < this->Ny; j++)
	  for (int i = 0; i < this->Nx; i++)
	    this->DryDepositionFlux(s, j, i) = 0.5
	      * (this->DepositionVelocity_i(s, j, i)
		 + this->DepositionVelocity_f(s, j, i))
	      * this->Concentration(this->DepositionVelocityGlobalIndex(s),
				    0, j, i);
    if (this->option_process["collect_dry_flux_aer"])
      for (int s = 0; s < this->Ns_aer; s++)
	for (int b = 0; b < Nbin_dep_aer; b++)
	  for (int j = 0; j < this->Ny; j++)
	    for (int i = 0; i < this->Nx; i++)
	      DryDepositionFlux_aer(s, b, j, i) = 0.5
		* (DepositionVelocity_aer_i(b, j, i)
		   + DepositionVelocity_aer_f(b, j, i))
		* this->Concentration_aer(s, bin_list_dep_aer[b], 0, j, i);

    if (this->option_process["with_diffusion"])
      this->Diffusion();

    if (this->option_process["with_point_emission"])
      this->PointEmission();
    if (this->option_process["with_point_emission_aer"])
      this->PointEmission_aer();
    if (this->option_process["with_volume_emission"]
	&& !this->option_process["with_chemistry"])
      for (int s = 0; s < this->Ns; s++)
	{
	  int k, j, i, emis_s;
	  if (this->HasVolumeEmission(s))
	    {
	      emis_s = this->VolumeEmissionIndex(s);
	      for (k = 0; k < this->Nz_vol_emis; k++)
		for (j = 0; j < this->Ny; j++)
		  for (i = 0; i < this->Nx; i++)
		    this->Concentration(s, k, j, i) +=
		      this->Delta_t * this->VolumeEmission_i(emis_s, k, j, i);
	    }
	}
    if (this->option_process["with_volume_emission_aer"]
	&& !this->option_process["with_chemistry"])
      for (int s = 0; s < this->Ns_aer; s++)
	  {
	    int b, k, j, i, emis_s, emis_b;
	    emis_b=0;
	    for (b = 0; b < this->Nbin_aer; b++)
	      if (this->HasVolumeEmission_aer(s,b))
		{
		  emis_s = VolumeEmissionIndex_aer(s);
		  for (k = 0; k < Nz_vol_emis_aer; k++)
		    for (j = 0; j < this->Ny; j++)
		      for (i = 0; i < this->Nx; i++)
			this->Concentration_aer(s, b, k, j, i) +=
			  this->Delta_t *
			  VolumeEmission_aer_i(emis_s, emis_b, k, j, i);
		  emis_b++;
		}
	  }

    if (this->option_process["collect_wet_flux"])
      this->WetDepositionFlux.SetZero();
    if (this->scavenging_model != "none")
      for (int s = 0; s < this->Ns_scav; s++)
 	{
	  T scavenging_ratio;
	  T cloud_height_mean;
 	  for (int j = 0; j < this->Ny; j++)
 	    for (int i = 0; i < this->Nx; i++)
 	      if (this->Rain_i(j, i) > 0.)
 		{
		  cloud_height_mean = 0.5
		    * (this->CloudHeight_i(j, i) + this->CloudHeight_f(j, i));
 		  for (int k = 0; k < this->Nz; k++)
		    if (cloud_height_mean > this->GridZ4D(k))
		      {
			scavenging_ratio =
			  exp(- 0.5 * this->Delta_t
			      * (this->ScavengingCoefficient_i(s, k, j, i)
				 + this->ScavengingCoefficient_f(s,
								 k, j, i)));
			if (this->option_process["collect_wet_flux"])
			  this->WetDepositionFlux(s, j, i) +=
			    this->Concentration
			    (this->ScavengingGlobalIndex(s), k, j, i)
			    / this->Delta_t * (1.- scavenging_ratio)
			    * (this->GridZ4D_interf(k+1)
			       - this->GridZ4D_interf(k));
			this->Concentration(this->ScavengingGlobalIndex(s),
					    k, j, i) *= scavenging_ratio;
		      }
		  if (this->option_process["collect_wet_flux"])
		    this->WetDepositionFlux(s, j, i) +=
		      this->InCloudWetDepositionFlux(s, j, i);
 		}
 	}

    if (this->option_process["collect_wet_flux_aer"])
      this->WetDepositionFlux_aer.SetZero();
    if (this->option_process["with_scavenging_aer"])
      for (int s = 0; s < this->Ns_aer; s++)
	for (int b = 0; b < Nbin_scav_aer; b++)
	  {
	    T scavenging_ratio;
	    for (int j = 0; j < this->Ny; j++)
	      for (int i = 0; i < this->Nx; i++)
		if (this->Rain_i(j, i) > 0.)
		  {
		    for (int k = 0; k < this->Nz; k++)
		      {
			scavenging_ratio =
			  exp(- 0.5 * this->Delta_t
			      * (ScavengingCoefficient_aer_i(b, k, j, i)
				 + ScavengingCoefficient_aer_f(b, k, j, i)));
			if (this->option_process["collect_wet_flux_aer"])
			  WetDepositionFlux_aer(s, b, j, i) +=
			    this->Concentration_aer(s, bin_list_scav_aer[b],
					      k, j, i)
			    / this->Delta_t * (1.- scavenging_ratio)
			    * (this->GridZ4D_interf(k+1)
			       - this->GridZ4D_interf(k));
			this->Concentration_aer(s, bin_list_scav_aer[b], k, j, i) *=
			  scavenging_ratio;
		      }
		    if (this->option_process["collect_wet_flux_aer"])
		      WetDepositionFlux_aer(s, b, j, i) +=
			InCloudWetDepositionFlux_aer(s, b, j, i);
		  }
	  }

    if (this->source_splitting && this->option_process["with_chemistry"])
      {
	this->Source_i.GetArray() =
	  (this->Concentration.GetArray() - this->Source_f.GetArray())
	  / this->Delta_t;
	this->Concentration.GetArray() = this->Source_f.GetArray();
	this->Source_f.GetArray() = this->Source_i.GetArray();
	
	if (this->option_process["with_volume_emission"])
	  for (int s = 0; s < this->Ns_vol_emis; s++)
	    {
	      int gs = this->VolumeEmissionGlobalIndex(s);
	      for (k = 0; k < this->Nz_vol_emis; k++)
		for (j = 0; j < this->Ny; j++)
		  for (i = 0; i < this->Nx; i++)
		    {
		      this->Source_i(gs, k, j, i)
			+= this->VolumeEmission_i(s, k, j, i);
		      this->Source_f(gs, k, j, i)
			+= this->VolumeEmission_f(s, k, j, i);
		    }
	    }

	this->Source_aer_i.GetArray() =
	  (this->Concentration_aer.GetArray() - this->Source_aer_f.GetArray())
	  / this->Delta_t;
	this->Concentration_aer.GetArray() = this->Source_aer_f.GetArray();
	this->Source_aer_f.GetArray() = this->Source_aer_i.GetArray();
	
	if (this->option_process["with_volume_emission_aer"])
	  for (int s = 0; s < Ns_vol_emis_aer; s++)
	    {
	      int gs = this->VolumeEmissionGlobalIndex_aer(s);
	      for (int b = 0; b < Nbin_vol_emis_aer; b++)
		{
		  int gb = species_list_vol_emis_aer[s].second[b];
		  for (k = 0; k < Nz_vol_emis_aer; k++)
		    for (j = 0; j < this->Ny; j++)
		      for (i = 0; i < this->Nx; i++)
			{
			  this->Source_aer_i(gs, gb, k, j, i)
			    += this->VolumeEmission_aer_i(s, b, k, j, i);
			  this->Source_aer_f(gs, gb, k, j, i)
			    += this->VolumeEmission_aer_f(s, b, k, j, i);
			}
		}
	    }
      }
    
    if (this->option_process["with_chemistry"])
      Chemistry();

    this->AddTime(this->Delta_t);
    this->step++;
  }


  ////////////////////
  // ACCESS METHODS //
  ////////////////////


 //! Returns the number of aerosol species with volume sources.
  /*! If source splitting is used, the number of species with volume sources
    is the total number of species; otherwise, it is the number of volume
    emissions.
    \return The number of species with volume sources.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::GetNs_source_aer()
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return this->Ns_aer;
    else
      return Ns_vol_emis_aer;
  }


  //! Returns the number of levels of aerosol volume sources.
  /*! If source splitting is used, the number of levels of volume sources
    is the total number of vertical layers; otherwise, it is the number of
    levels in volume emissions.
    \return The number of levels of volume sources.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::GetNz_source_aer()
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return this->Nz;
    else
      return Nz_vol_emis_aer;
  }


  //! Returns the number of bins in aerosol volume sources.
  /*! \return The maximum number of bins of aerosol volume sources.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::GetNbinMax_source_aer()
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return this->Nbin_aer;
    else
      return Nbin_vol_emis_aer;
  }


  //! Checks whether an aerosol has sources.
  /*! If source splitting is used, any aerosol has sources; otherwise, it
    checks whether the aerosol has volume emissions.
    \param s species global index.
    \param b bin number.
    \return True if the aerosol has sources, false otherwise.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  bool Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::HasSource_aer(int s, int b)
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return true;
    else
      return HasVolumeEmission_aer(s, b);
  }


  //! Returns the global index of an aerosol with volume sources.
  /*!
    \param s species index in volume sources.
    \return The species global index.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::SourceGlobalIndex_aer(int s)
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return s;
    else
      return VolumeEmissionGlobalIndex_aer(s);
  }


  /*! \brief Returns the global index of an aerosol bin with volume sources
    for a given species. */
  /*!
    \param s species index in volume sources.
    \param b bin index in volume sources.
    \return The bin global index for the given species.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  int Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::SourceGlobalBinIndex_aer(int s, int b)
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return b;
    else
      return this->species_list_vol_emis_aer[s].second[b];
  }


  //! Returns the aerosol sources at current date.
  /*!
    \return The aerosol sources at current date.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  Data<T, 5>&
  Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::GetSource_aer_i()
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return Source_aer_i;
    else
      return VolumeEmission_aer_i;
  }


  //! Returns the aerosol sources at next date.
  /*!
    \return The aerosol sources at next date.
   */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  Data<T, 5>&
  Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::GetSource_aer_f()
  {
    if (this->source_splitting && this->option_process["with_chemistry"])
      return Source_aer_f;
    else
      return VolumeEmission_aer_f;
  }


  ///////////////////////
  // PROTECTED METHODS //
  ///////////////////////


  //! Moves model input-data to the current date.
  /*! This method prepares the model for a time integration from the current
    date. It reads input data to related to chemistry be read before InitStep
    and Forward.
  */
  template<class T, class ClassAdvection,
	   class ClassDiffusion, class ClassChemistry>
  void Polair3DAerosol<T, ClassAdvection, ClassDiffusion, ClassChemistry>
  ::InitAllData()
  {
    int k, i, j;
    string species, species_bin;
    vector<int> bin;
    string filename;
    Date date;
    T Delta_t;

    Polair3DChemistry<T, ClassAdvection,
      ClassDiffusion, ClassChemistry>::InitAllData();

    /*** Surface temperature ***/

    if (this->option_manage["surface_temperature"])
      this->InitData("meteo", "SurfaceTemperature",
		     FileSurfaceTemperature_i, FileSurfaceTemperature_f,
		     this->current_date, SurfaceTemperature_f);

    /*** Surface pressure ***/

    if (this->option_manage["surface_pressure"])
      this->InitData("meteo", "SurfacePressure",
		     FileSurfacePressure_i, FileSurfacePressure_f,
		     this->current_date, SurfacePressure_f);

    /*** First level wind module ***/

    T zonal_wind, meridional_wind;
    if (this->option_manage["first_level_wind_module"])
      for (j = 0; j < this->Ny; j++)
	for (i = 0; i < this->Nx; i++)
	  {
	    zonal_wind = 0.5 * (this->ZonalWind_i(0, j, i)
				+ this->ZonalWind_i(0, j, i + 1));
	    meridional_wind = 0.5 * (this->MeridionalWind_i(0, j, i)
				     + this->MeridionalWind_i(0, j + 1, i));
	    FirstLevelWindModule_f(j, i) =
	      sqrt(zonal_wind * zonal_wind
		   + meridional_wind * meridional_wind);
	  }

    /*** Liquid water content ***/

    if (this->option_manage["liquid_water_content"])
      this->InitData("meteo", "LiquidWaterContent",
		     FileLiquidWaterContent_i, FileLiquidWaterContent_f,
		     this->current_date, LiquidWaterContent_i);


    /*** Relative humidity ***/
    
    for (k = 0; k < this->Nz; k++)
      for (j = 0; j < this->Ny; j++)
	for (i = 0; i < this->Nx; i++)
	  {
	    _compute_relative_humidity(&(this->SpecificHumidity_f(k, j, i)),
				       &(this->Temperature_f(k, j, i)),
				       &(this->Pressure_f(k, j, i)),
				       &RelativeHumidity_f(k, j, i));
	    RelativeHumidity_f(k, j, i) =
	      min(max(RelativeHumidity_f(k, j, i), 0.), 1.);
	  }

    /*** Aerosol wet diameter ***/

    this->InitWetDiameter_aer(RelativeHumidity_f, this->Temperature_f,
			      WetDiameter_aer);

    /*** Snow height ***/

    if (this->option_manage["snow_height"])
      this->InitData("meteo", "SnowHeight", FileSnowHeight_i,
		     FileSnowHeight_f, this->current_date, SnowHeight_f);

    /*** Boundary conditions ***/

    if (this->option_manage["boundary_condition_aer"])
      for (i = 0; i < Ns_bc_aer; i++)
	{
	  species = species_list_bc_aer[i].first;
	  bin = species_list_bc_aer[i].second;
	  for (j = 0; j < int(bin.size()); j++)
	    {
	      species_bin = species + string("_") + to_str(bin[j]);
	      string filename
		= this->input_files["boundary_condition_aer"](species_bin);
	      date = this->input_files["boundary_condition_aer"].GetDateMin();
	      Delta_t
		= this->input_files["boundary_condition_aer"].GetDelta_t();
	      this->InitData(find_replace(filename, "&c", "z"), date, Delta_t,
			     FileBoundaryCondition_z_aer_i,
			     FileBoundaryCondition_z_aer_f,
			     this->current_date, i, j,
			     BoundaryCondition_z_aer_i);
	      this->InitData(find_replace(filename, "&c", "y"), date, Delta_t,
			     FileBoundaryCondition_y_aer_i,
			     FileBoundaryCondition_y_aer_f,
			     this->current_date, i, j,
			     BoundaryCondition_y_aer_i);
	      this->InitData(find_replace(filename, "&c", "x"), date, Delta_t,
			     FileBoundaryCondition_x_aer_i,
			     FileBoundaryCondition_x_aer_f,
			     this->current_date, i, j,
			     BoundaryCondition_x_aer_i);
	    }
	}

    /*** Deposition ***/

    if (this->option_manage["deposition_velocity_aer"])
      if (!this->option_process["compute_deposition_aer"])
	for (i = 0; i < Nbin_dep_aer; i++)
	  this->InitData("deposition_velocity_aer",
			 to_str(bin_list_dep_aer[i]),
			 FileDepositionVelocity_aer_i,
			 FileDepositionVelocity_aer_f,
			 this->current_date, i, DepositionVelocity_aer_f);
      else
	this->InitDepositionVelocity(this->Temperature_f, this->Pressure_f,
				     this->SurfaceTemperature_f,
				     this->SurfacePressure_f,
				     FirstLevelWindModule_f,
				     RelativeHumidity_f, SnowHeight_f,
				     DepositionVelocity_aer_f);

    /*** Scavenging coefficients ***/

    if (this->option_manage["scavenging_coefficient"]
	&& this->scavenging_model == "microphysical-ph")
      InitScavengingCoefficient(this->Temperature_f, this->Pressure_f,
				LiquidWaterContent_i, pH,
				this->CloudHeight_f, this->Rain_i,
				this->ScavengingCoefficient_f);
 
    if (this->option_manage["scavenging_coefficient_aer"]
	&& this->option_process["with_scavenging_aer"])
      this->InitScavengingCoefficient_aer(this->Temperature_f,
					  this->Pressure_f,
					  WetDiameter_aer,
					  this->CloudHeight_f,
					  this->Rain_i,
					  ScavengingCoefficient_aer_f);

    /*** Surface emissions ***/

    if (this->option_manage["surface_emission_aer"])
      for (i = 0; i < Ns_surf_emis_aer; i++)
	{
	  species = species_list_surf_emis_aer[i].first;
	  bin = species_list_surf_emis_aer[i].second;
	  for (j = 0; j < int(bin.size()); j++)
	    {
	      species_bin = species + string("_") + to_str(bin[j]);
	      this->InitData("surface_emission_aer", species_bin,
			     FileSurfaceEmission_aer_i,
			     FileSurfaceEmission_aer_f,
			     this->current_date, i, j,
			     SurfaceEmission_aer_f);
	    }
	}

    /*** Volume emissions ***/

    if (this->option_manage["volume_emission_aer"])
      for (i = 0; i < Ns_vol_emis_aer; i++)
	{
	  species = species_list_vol_emis_aer[i].first;
	  bin = species_list_vol_emis_aer[i].second;
	  for (j = 0; j < int(bin.size()); j++)
	    {
	      species_bin = species + string("_") + to_str(bin[j]);
	      this->InitData("volume_emission_aer", species_bin,
			     FileVolumeEmission_aer_i,
			     FileVolumeEmission_aer_f,
			     this->current_date, i, j,
			     VolumeEmission_aer_f);
	    }
	}
  }


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODELS_POLAIR3DAEROSOL_CXX
#endif
