// Copyright (C) 2006-2007 CEREA
//     Author(s): Edouard Debry
//
// CEREA (http://www.enpc.fr/cerea/) is a joint laboratory of ENPC
// (http://www.enpc.fr/) and EDF R&D (http://www.edf.fr/).
//
// This file is part of the Size Resolved Aerosol Module,
// which is a component of the air quality modeling system Polyphemus. 
//
// Polyphemus is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// Polyphemus is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
// For more information, please see the Polyphemus web site:
//     http://www.enpc.fr/cerea/polyphemus/


#ifndef POLYPHEMUS_FILE_MODULES_AEROSOL_AEROSOLRACM_SIREAM_HXX


#include <vector>
#include "AtmoData.hxx"
#include "BaseModule.cxx"

namespace Polyphemus
{


  using namespace std;
  using namespace AtmoData;


  //////////////////////
  // FORTRAN FUNCTION //
  //////////////////////

#ifdef POLYPHEMUS_SINGLE_UNDERSCORE
#undef POLYPHEMUS_DOUBLE_UNDERSCORE
#elif defined(__GNUG__) && __GNUG__ < 4 && !defined(__INTEL_COMPILER)
#undef POLYPHEMUS_DOUBLE_UNDERSCORE
#define POLYPHEMUS_DOUBLE_UNDERSCORE
#endif

#ifdef POLYPHEMUS_DOUBLE_UNDERSCORE
#define _compute_coagulation_coefficient compute_coagulation_coefficient__
#else
#define _compute_coagulation_coefficient compute_coagulation_coefficient_
#endif

#define _chem chem_
#define _aerosol aerosol_

  extern "C"
  {
    void _chem(int*, int*, int*, int*, int*, int*, int*, int*, int*, int*,
	       double*, double*, double*, double*, double*, double*,
	       double*, double*, double*, double*, double*,
	       double*, double*, double*, double*, double*, int*,
	       double*, double*, double*, int*, int*, int*, int*,
	       double*, double*, double*, double*, int*, double*);

    void _aerosol(int*, int*, int*, int*, double*, double*, double*,
		  double*, double*, double*, double*, int*, int*,
		  int*, int*, int*, double*, double*, double*, double*,
		  double*, int*, int*, int*, double*, double*, double*,
		  double*, double*, int*, int*, int*, int*, int*,
		  double*, double*);
    
    void _compute_coagulation_coefficient(int*, double*,
					  int*, int*, int*, double*);
  }


  //////////////////////////
  // AEROSOLRACM_SIREAM //
  //////////////////////////


  //! This class is a numerical solver for the chemical mechanism RACM_SIREAM.
  /*! It uses a second-order Rosenbrock method.
   */
  template<class T>
  class AerosolRACM_SIREAM: public BaseModule
  {

  protected:

    //! Number of species.
    int Ns;
    //! Number of reactions.
    int Nr;
    //! Number of photolysis reactions.
    int Nr_photolysis;
    //! Number of species with volume sources.
    int Ns_source;
    //! Number of levels with volume sources.
    int Nz_source;

    //! Number of sub-cycles.
    int Ncycle;

    //! Sorted list of species names.
    vector<string> species_list;
    
    //! Molecular weights of species.
    Array<T, 1> molecular_weight;
    /*! \brief Map between photolysis reactions names and their indices in
      photolysis reactions.
     */
    map<string, int> photolysis_reaction_name;
    //! Indices of photolysis reactions among other reactions.
    Array<int, 1> photolysis_reaction_index;
    //! Indices of species with volume sources.
    Array<int, 1> source_index;

    //! Number of inorganic aerosol species.
    int Ns_inorganic_aer;
   
    //! Number of inert aerosol species.
    int Ns_inert_aer;
    
    //! Number of organic aerosol species.
    int Ns_organic_aer;
    
    //! Number of sub-cycles ofr aerosol processes
    int Ncycle_aer;
   
    //! Number of aerosol species.
    int Ns_aer;
 
    //! Number of aerosol bins.
    int Nbin_aer;

    //! Bin index corresponding to fixed cutting diameter.
    int cutting_bin;
    /* bins from 1 to cutting_bin included are at equilibrium
       and are dynamic from cutting_bin+1 to Nbin_aer */

    //! Indices of species with heterogeneous reactions.
    Array<int, 1> heterogeneous_reaction_index;

    //! list of aerosol options.
    int Noptions_aer;
    Array<int, 1> options_aer;

    //! Number of species with volume sources.
    int Ns_source_aer;
    //! Maximum number of bins with volume sources.
    int NbinMax_source_aer;
    //! Number of levels with volume sources.
    int Nz_source_aer;

    //! Indices of species with volume sources.
    Array<int, 1> source_index_species_aer;

    //! Indices of bins with volume sources for each species.
    Array<int, 2> source_index_bin_aer;

  public:
    /*** Configuration ***/

    map<string, bool> option_process_aer;
   
    //! Numerical solver for dynamic bin condensation (etr, ros2 or ebi).
    string dynamic_condensation_solver;

    //! Cutting diameter between equilibrium and dynamic bins.
    double fixed_cutting_diameter;

    //! Sulfate condensation computation method (equilibrium, dynamic).
    string sulfate_computation;

    /*! Redistribution method of lagrangian bins (number-conserving,
      interpolation).
    */
    string redistribution_method;

    //! If nucleation, which nucleation model (binary, ternary).
    string nucleation_model;

    //! Aerosol wet diameter estimation (gerber or isoropia).
    string wet_diameter_estimation;

    //! Bins bounds (in m).
    Array<T, 1> BinBound_aer;

    //! Aerosol density (kg / m^3).
    double FixedDensity_aer;
    Array<T, 1> Density_aer;

    //! Coagulation coefficients.
    Array<int, 1> CoagulationCouple;
    Array<int, 2> CoagulationFirstIndex;
    Array<int, 2> CoagulationSecondIndex;
    Array<T, 3> CoagulationPartitionCoefficient;

    //! liquid water content threshold for clouds.
    double lwc_cloud_threshold;

    /*** Constructor ***/

    AerosolRACM_SIREAM();

    /*** Other methods ***/

    template<class ClassModel>
    void Init(ClassModel& Model);
    
    template<class ClassModel>
    void Forward(ClassModel& Model);

    template<class ClassModel>
    void Forward_aer(ClassModel& Model);

    void Forward(T current_time,
		 Data<T, 3>& Attenuation_i,
		 Data<T, 3>& SpecificHumidity_i,
		 Data<T, 3>& Temperature_i,
		 Data<T, 3>& Pressure_i,
		 Data<T, 4>& VolumeEmission_i,
		 Data<T, 4>& PhotolysisRate_i,
		 T next_time,
		 Data<T, 3>& Attenuation_f,
		 Data<T, 3>& SpecificHumidity_f,
		 Data<T, 3>& Temperature_f,
		 Data<T, 3>& Pressure_f,
		 Data<T, 4>& VolumeEmission_f,
		 Data<T, 4>& PhotolysisRate_f,
		 Array<T, 1>& Longitude,
		 Array<T, 1>& Latitude,
		 Data<T, 4>& Concentration,
		 Data<T, 3>& LiquidWaterContent_i,
		 Data<T, 4>& WetDiameter_aer,
		 Data<T, 5>& Concentration_aer);

    void Forward_aer(T current_time,
		 Data<T, 3>& SpecificHumidity_i,
		 Data<T, 3>& Temperature_i,
		 Data<T, 3>& Pressure_i,
		 T next_time,
		 Data<T, 4>& Concentration,
		 Data<T, 3>& LiquidWaterContent_i,
		 Data<T, 2>& Rain_i,
		 Array<T, 1>& VerticalInterface,
		 Data<T, 5>& Concentration_aer,
		 Data<T, 3>& InCloudWetDepositionFlux,
		 Data<T, 4>& InCloudWetDepositionFlux_aer,
		 Data<T, 5>& Source_aer_i,
		 Data<T, 3>& pH);

    bool IsRequired(string field);
    bool IsComputed(string field);
  };


} // namespace Polyphemus.


#define POLYPHEMUS_FILE_MODULES_AEROSOL_AEROSOLRACM_SIREAM_HXX
#endif
